"""Parser for Acacia."""

__all__ = ['Parser']

from typing import (
    Callable, Optional, List, Tuple, NamedTuple, Type, TYPE_CHECKING
)

from acaciamc.error import *
from acaciamc.tokenizer import TokenType, BRACKETS
from acaciamc.ast import *

if TYPE_CHECKING:
    from acaciamc.tokenizer import Tokenizer

class FuncType(NamedTuple):
    allowed_ports: Tuple[FuncPortType, ...]
    type_required: bool
    ast_class: Type[FuncData]

FUNC_NORMAL = FuncType(
    allowed_ports=(FuncPortType.by_value,),
    type_required=True,
    ast_class=NormalFuncData
)
FUNC_INLINE = FuncType(
    allowed_ports=(FuncPortType.by_value, FuncPortType.by_reference,
                   FuncPortType.const),
    type_required=False,
    ast_class=InlineFuncData
)
FUNC_CONST = FuncType(
    allowed_ports=(FuncPortType.by_value,),
    type_required=False,
    ast_class=ConstFuncData
)

class Parser:
    def __init__(self, tokenizer: "Tokenizer"):
        self.tokenizer = tokenizer
        self.current_token = self.tokenizer.get_next_token()
        self.next_token = None

    @property
    def current_pos(self):
        return {
            "lineno": self.current_token.lineno,
            "col": self.current_token.col
        }

    def error(self, err_type: ErrorType, lineno=None, col=None, **kwargs):
        if lineno is None:
            lineno = self.current_pos['lineno']
        if col is None:
            col = self.current_pos['col']
        err = Error(err_type, **kwargs)
        err.location.linecol = (lineno, col)
        raise err

    def eat(self, expect_token_type: Optional[TokenType] = None):
        """Move to next token.
        If `expect_token_type` is given, also check type of the old
        token.
        """
        if ((expect_token_type is not None) and
            (self.current_token.type != expect_token_type)):
            self.error(ErrorType.UNEXPECTED_TOKEN, token=self.current_token)
        if self.next_token is None:
            self.current_token = self.tokenizer.get_next_token()
        else:
            self.current_token = self.next_token
            self.next_token = None

    def peek(self):
        """Peek the next token and store at `self.next_token`."""
        if self.next_token is None:
            # only generate when not generated yet
            self.next_token = self.tokenizer.get_next_token()

    def _block(self, func: Callable[[], Statement]):
        """
        Read an indented block of structures.
        `func` reads the structure and should return the result AST.
        block{n} := NEW_LINE INDENT n+ DEDENT
        """
        stmts = []
        self.eat(TokenType.new_line)
        if self.current_token.type is not TokenType.indent:
            self.error(ErrorType.EMPTY_BLOCK)
        self.eat()  # INDENT
        while self.current_token.type != TokenType.dedent:
            stmts.append(func())
        self.eat()  # DEDENT
        return stmts

    def statement_block(self):
        """
        Read a block of statements.
        statement_block := block{statement}
        """
        return self._block(self.statement)

    # Following are different AST generators
    ## Expression generator

    def literal(self):
        """literal := INTEGER | TRUE | FALSE | NONE | FLOAT"""
        pos = self.current_pos
        tok_type = self.current_token.type
        if tok_type in (TokenType.integer, TokenType.float_):
            value = self.current_token.value
            self.eat()
        elif tok_type in (TokenType.true, TokenType.false):
            value = tok_type is TokenType.true
            self.eat()
        elif tok_type is TokenType.none:
            value = None
            self.eat()
        else:
            self.error(ErrorType.UNEXPECTED_TOKEN, token=self.current_token)
        return Literal(value, **pos)

    def str_literal(self):
        """str_literal := (STRING_BEGIN formatted_str STRING_END)+"""
        pos = self.current_pos
        content: List[str] = []
        while self.current_token.type is TokenType.string_begin:
            self.eat()  # eat STRING_BEGIN
            content.extend(self.formatted_str().content)
            self.eat(TokenType.string_end)
        return StrLiteral(FormattedStr(content, **pos), **pos)

    def identifier(self):
        """identifier := IDENTIFIER"""
        pos = self.current_pos
        token = self.current_token
        self.eat(TokenType.identifier)
        return Identifier(token.value, **pos)

    def self(self):
        """self := SELF"""
        pos = self.current_pos
        self.eat(TokenType.self)
        return Self(**pos)

    def list_or_map(self):
        """
        list := paren_list_of{LBRACE, expr, RBRACE}
        map_item := expr COLON expr
        map := LBRACE ((map_item (COMMA map_item)* COMMA?) | COLON)
            RBRACE
        """
        pos = self.current_pos
        self.eat(TokenType.lbrace)
        # Check for empty list or map
        if self.current_token.type is TokenType.rbrace:
            self.eat()
            return ListDef(items=[], **pos)
        elif self.current_token.type is TokenType.colon:
            self.eat()
            self.eat(TokenType.rbrace)
            return MapDef(keys=[], values=[], **pos)
        first = self.expr()
        if self.current_token.type is TokenType.colon:
            # Map
            self.eat()
            keys = [first]
            values = [self.expr()]
            if self.current_token.type is TokenType.comma:
                self.eat()
                while self.current_token.type is not TokenType.rbrace:
                    keys.append(self.expr())
                    self.eat(TokenType.colon)
                    values.append(self.expr())
                    if self.current_token.type is TokenType.comma:
                        self.eat()
                    else:
                        break
            self.eat(TokenType.rbrace)
            return MapDef(keys, values, **pos)
        else:
            # List
            items = [first]
            if self.current_token.type is TokenType.comma:
                self.eat()
                while self.current_token.type is not TokenType.rbrace:
                    items.append(self.expr())
                    if self.current_token.type is TokenType.comma:
                        self.eat()
                    else:
                        break
            self.eat(TokenType.rbrace)
            return ListDef(items, **pos)

    def expr_l0(self):
        """
        expr_l0 := (LPAREN expr RPAREN) | literal | identifier
            | str_literal | self | list | map
        """
        if self.current_token.type in (
            TokenType.integer, TokenType.float_, TokenType.true,
            TokenType.false, TokenType.none
        ):
            return self.literal()
        elif self.current_token.type is TokenType.identifier:
            return self.identifier()
        elif self.current_token.type is TokenType.string_begin:
            return self.str_literal()
        elif self.current_token.type is TokenType.lparen:
            self.eat(TokenType.lparen)
            node = self.expr()
            self.eat(TokenType.rparen)
            return node
        elif self.current_token.type is TokenType.self:
            return self.self()
        elif self.current_token.type is TokenType.lbrace:
            return self.list_or_map()
        else:
            self.error(ErrorType.UNEXPECTED_TOKEN, token=self.current_token)

    def expr_l1(self):
        """
        expr_l1 := expr_l0 (
            (POINT IDENTIFIER)
            | call_table
            | paren_list_of{LBRACKET, expr, RBRACKET}
        )*
        """
        node = self.expr_l0()
        def _attribute(node: Expression):
            self.eat(TokenType.point)
            attr = self.current_token.value
            self.eat(TokenType.identifier)
            return Attribute(node, attr, lineno=node.lineno, col=node.col)

        def _call(node: Expression):
            return Call(
                node, self.call_table(),
                lineno=node.lineno, col=node.col
            )

        def _subscript(node: Expression):
            subscripts = []
            self.paren_list_of(
                lambda: subscripts.append(self.expr()),
                lparen=TokenType.lbracket
            )
            return Subscript(
                node, subscripts, lineno=node.lineno, col=node.col
            )

        # start
        while True:
            if self.current_token.type is TokenType.point:
                self.peek()
                if self.next_token.type is TokenType.new:
                    # If we find `new` then stop parsing the expression
                    # `statement` is responsible for parsing that
                    break
                node = _attribute(node)
            elif self.current_token.type is TokenType.lparen:
                node = _call(node)
            elif self.current_token.type is TokenType.lbracket:
                node = _subscript(node)
            else:
                break
        return node

    def expr_l2(self):
        """expr_l2 := ((PLUS | MINUS) expr_l2) | expr_l1"""
        pos = self.current_pos
        if self.current_token.type is TokenType.plus:
            self.eat()
            return UnaryOp(Operator.positive, self.expr_l2(), **pos)
        elif self.current_token.type is TokenType.minus:
            self.eat()
            return UnaryOp(Operator.negative, self.expr_l2(), **pos)
        else:  # no unary operators
            return self.expr_l1()

    def expr_l3(self):
        """expr_l3 := expr_l2 ((STAR | SLASH | MOD) expr_l2)*"""
        node = self.expr_l2()
        while True:
            token_type = self.current_token.type
            if token_type is TokenType.star:
                op = Operator.multiply
            elif token_type is TokenType.slash:
                op = Operator.divide
            elif token_type is TokenType.mod:
                op = Operator.mod
            else:  # no valid operator found
                return node
            self.eat()  # eat operator
            node = BinOp(
                node, op, self.expr_l2(),
                lineno=node.lineno, col=node.col
            )

    def expr_l4(self):
        """expr_l4 := expr_l3 ((ADD | MINUS) expr_l3)*"""
        node = self.expr_l3()
        while True:
            token_type = self.current_token.type
            if token_type is TokenType.plus:
                op = Operator.add
            elif token_type is TokenType.minus:
                op = Operator.minus
            else:  # no valid operator found
                return node
            self.eat()  # eat operator
            node = BinOp(node, op, self.expr_l3(),
                         lineno=node.lineno, col=node.col)

    def expr_l5(self):
        """
        expr_l5 := expr_l4 ((EQUAL_TO | UNEQUAL_TO | GREATER | LESS
            | GREATER_EQUAL | LESS_EQUAL) expr_l4)*
        """
        pos = self.current_pos
        left = self.expr_l4()
        COMPARE_OPS = (
            TokenType.equal_to,
            TokenType.unequal_to,
            TokenType.greater,
            TokenType.less,
            TokenType.greater_equal,
            TokenType.less_equal
        )
        operands, operators = [], []
        while self.current_token.type in COMPARE_OPS:
            operators.append(Operator[self.current_token.type.name])
            self.eat()
            operands.append(self.expr_l4())
        if operators:  # if not empty
            return CompareOp(left, operators, operands, **pos)
        return left

    def expr_l6(self):
        """expr_l6 := (NOT expr_l6) | expr_l5"""
        pos = self.current_pos
        if self.current_token.type is TokenType.not_:
            self.eat()
            return UnaryOp(Operator.not_, self.expr_l6(), **pos)
        else:  # no unary operators
            return self.expr_l5()

    def expr_l7(self):
        """expr_l7 := expr_l6 (AND expr_l6)*"""
        left = self.expr_l6()
        operands = []
        while self.current_token.type is TokenType.and_:
            self.eat()  # eat and_
            operands.append(self.expr_l6())
        if operands:  # if not empty
            operands.insert(0, left)
            return BoolOp(Operator.and_, operands,
                          lineno=left.lineno, col=left.col)
        return left

    def expr_l8(self):
        """expr_l8 := expr_l7 (OR expr_l7)*"""
        left = self.expr_l7()
        operands = []
        while self.current_token.type is TokenType.or_:
            self.eat()  # eat or_
            operands.append(self.expr_l7())
        if operands:  # if not empty
            operands.insert(0, left)
            return BoolOp(Operator.or_, operands,
                          lineno=left.lineno, col=left.col)
        return left

    # expr: keep updates with the highest level of expr method
    # this is to make sure other funcs always call the
    # highest level of expr (convenient when updating)
    expr = expr_l8

    ## Statement generator

    def if_stmt(self):
        """
        if_stmt := IF expr COLON statement_block
            (ELIF expr COLON statement_block)*
            (ELSE COLON statement_block)?
        """
        pos = self.current_pos
        def _if_extra():
            """
            if_extra := (ELIF expr COLON statement_block if_extra?)
                | (ELSE COLON statement_block)
            return list of statements
            """
            if self.current_token.type is TokenType.else_:
                self.eat()
                self.eat(TokenType.colon)
                return self.statement_block()
            elif self.current_token.type is TokenType.elif_:
                elif_pos = self.current_pos
                self.eat()
                condition = self.expr()
                self.eat(TokenType.colon)
                stmts = self.statement_block()
                # See if there are more "elif" or "else"
                else_stmts = _if_extra()
                return [If(condition, stmts, else_stmts, **elif_pos)]
            else:
                return []
        # if_stmt := IF expr COLON statement_block if_extra?
        self.eat(TokenType.if_)
        condition = self.expr()
        self.eat(TokenType.colon)
        stmts = self.statement_block()
        else_stmts = _if_extra()
        return If(condition, stmts, else_stmts, **pos)

    def while_stmt(self):
        """while_stmt := WHILE expr COLON statement_block"""
        pos = self.current_pos
        self.eat(TokenType.while_)
        condition = self.expr()
        self.eat(TokenType.colon)
        body = self.statement_block()
        return While(condition, body, **pos)

    def pass_stmt(self):
        """pass_stmt := PASS"""
        node = Pass(**self.current_pos)
        self.eat(TokenType.pass_)
        return node

    def interface_stmt(self):
        """
        interface_stmt := INTERFACE (INTERFACE_PATH | str_literal)
            COLON statement_block
        """
        pos = self.current_pos
        self.eat(TokenType.interface)
        if self.current_token.type is TokenType.interface_path:
            path = self.current_token.value
            self.eat()
        else:
            path = self.str_literal()
        self.eat(TokenType.colon)
        stmts = self.statement_block()
        return InterfaceDef(path, stmts, **pos)

    def _func_port_qualifier(self, allowed_ports: Tuple[FuncPortType]) \
            -> FuncPortType:
        """func_port_inline := (CONST | AMPERSAND)?"""
        pos = self.current_pos
        if self.current_token.type is TokenType.const:
            self.eat()
            res = FuncPortType.const
        elif self.current_token.type is TokenType.ampersand:
            self.eat()
            res = FuncPortType.by_reference
        else:
            res = FuncPortType.by_value
        if res not in allowed_ports:
            self.error(ErrorType.INVALID_FUNC_PORT, **pos, port=res.localized)
        return res

    def _def_head(self) -> str:
        """def_head := DEF IDENTIFIER"""
        self.eat(TokenType.def_)
        name = self.current_token.value
        self.eat(TokenType.identifier)
        return name

    def _function_def(self, t: FuncType) -> FuncData:
        pos = self.current_pos
        arg_table = self.argument_table(t.allowed_ports, t.type_required)
        if self.current_token.type is TokenType.arrow:
            self.eat()
            pos = self.current_pos
            qualifier = self._func_port_qualifier(t.allowed_ports)
            ret_type = self.type_spec()
            returns = FunctionPort(ret_type, qualifier, **pos)
        else:
            returns = None
        self.eat(TokenType.colon)
        stmts = self.statement_block()
        return t.ast_class(arg_table, stmts, returns, **pos)

    def const_def_stmt(self):
        """
        funcdef_const := argtable_const (ARROW type_spec)?
            COLON statement_block
        const_def_stmt := CONST def_head fundef_const
        """
        pos = self.current_pos
        self.eat(TokenType.const)
        name = self._def_head()
        func_data = self._function_def(FUNC_CONST)
        return FuncDef(name, func_data, **pos)

    def inline_def_stmt(self):
        """
        funcdef_inline := argtable_inline
            (ARROW func_port_inline type_spec)?
            COLON statement_block
        inline_def_stmt := INLINE def_head funcdef_inline
`        """
        pos = self.current_pos
        self.eat(TokenType.inline)
        name = self._def_head()
        func_data = self._function_def(FUNC_INLINE)
        return FuncDef(name, func_data, **pos)

    def def_stmt(self):
        """
        funcdef_normal := argtable_normal (ARROW type_spec)?
            COLON statement_block
        def_stmt := def_head funcdef_normal
        """
        pos = self.current_pos
        name = self._def_head()
        func_data = self._function_def(FUNC_NORMAL)
        return FuncDef(name, func_data, **pos)

    def _entity_body(self):
        pos = self.current_pos
        if self.current_token.type is TokenType.identifier:
            # entity_field_decl
            name = self.current_token.value
            self.eat()  # eat IDENTIFIER
            self.eat(TokenType.colon)
            type_ = self.type_spec()
            self.eat(TokenType.new_line)
            return EntityField(name, type_, **pos)
        elif self.current_token.type is TokenType.pass_:
            res = self.pass_stmt()
            self.eat(TokenType.new_line)
            return res
        else:
            # method_decl or new_method_decl
            if self.current_token.type is TokenType.const:
                self.peek()
                if self.next_token.type is TokenType.new:
                    self.error(ErrorType.CONST_NEW_METHOD, **pos)
            if self.current_token.type is TokenType.inline:
                self.peek()
                tok = self.next_token
            else:
                tok = self.current_token
            if tok.type is TokenType.new:
                if self.current_token.type is TokenType.inline:
                    self.eat()
                    t = FUNC_INLINE
                else:
                    t = FUNC_NORMAL
                self.eat(TokenType.new)
                data = self._function_def(t)
                return NewMethod(data, **pos)
            qualifier = MethodQualifier.none
            if self.current_token.type is TokenType.virtual:
                self.eat()  # eat "virtual"
                qualifier = MethodQualifier.virtual
            elif self.current_token.type is TokenType.override:
                self.eat()  # eat "override"
                qualifier = MethodQualifier.override
            elif self.current_token.type is TokenType.static:
                self.eat()  # eat "static"
                qualifier = MethodQualifier.static
            if self.current_token.type is TokenType.inline:
                content = self.inline_def_stmt()
            elif self.current_token.type is TokenType.const:
                if qualifier is not MethodQualifier.static:
                    self.error(ErrorType.NONSTATIC_CONST_METHOD, **pos)
                content = self.const_def_stmt()
            else:
                content = self.def_stmt()
            return EntityMethod(content, qualifier, **pos)

    def _extends_clause(self) -> List[Expression]:
        """extends_clause := (EXTENDS list_of{expr})?"""
        bases = []
        if self.current_token.type is TokenType.extends:
            self.eat()
            self.list_of(lambda: bases.append(self.expr()))
        return bases

    def entity_stmt(self):
        """
        entity_stmt := ENTITY IDENTIFIER extends_clause
            COLON block{entity_body}
        entity_field_decl := IDENTIFIER COLON type_spec
        method_decl := (
            (VIRTUAL | OVERRIDE | STATIC)? (def_stmt | inline_def_stmt)
            | (STATIC const_def_stmt)
        )
        new_method_decl := (NEW funcdef_normal)
            | (INLINE NEW funcdef_inline)
        entity_body := (method_decl | new_method_decl)
            | ((entity_field_decl | pass_stmt) NEW_LINE)
        """
        pos = self.current_pos
        self.eat(TokenType.entity)
        name = self.current_token.value
        self.eat(TokenType.identifier)
        parents = self._extends_clause()
        self.eat(TokenType.colon)
        body = self._block(self._entity_body)
        new_methods = [v for v in body if isinstance(v, NewMethod)]
        if len(new_methods) > 1:
            self.error(ErrorType.MULTIPLE_NEW_METHODS, **pos)
        elif new_methods:
            new_method = new_methods[0]
            body.remove(new_method)
        else:
            new_method = None
        return EntityTemplateDef(name, parents, body, new_method, **pos)

    def formatted_str(self):
        """formatted_str := (STRING_BODY | (DOLLAR_LBRACE expr RBRACE))*"""
        pos = self.current_pos
        res = []
        while self.current_token.type in (
            TokenType.text_body, TokenType.dollar_lbrace
        ):
            if self.current_token.type is TokenType.text_body:
                res.append(self.current_token.value)
                self.eat()
            else:
                self.eat(TokenType.dollar_lbrace)
                res.append(self.expr())
                self.eat(TokenType.rbrace)
        return FormattedStr(res, **pos)

    def command_stmt(self):
        """command_stmt := COMMAND_BEGIN formatted_str COMMAND_END"""
        pos = self.current_pos
        self.eat(TokenType.command_begin)
        content = self.formatted_str()
        self.eat(TokenType.command_end)
        return Command(content, **pos)

    def module_meta(self):
        """module_meta := POINT* IDENTIFIER (POINT IDENTIFIER)*"""
        # leading dots
        leadint_dots = 0
        while self.current_token.type is TokenType.point:
            leadint_dots += 1
            self.eat()
        # at least one name should be given
        names = [self.current_token.value]
        self.eat(TokenType.identifier)
        # read more names
        while self.current_token.type is TokenType.point:
            self.eat()
            names.append(self.current_token.value)
            self.eat(TokenType.identifier)
        last_name = names.pop()
        return ModuleMeta(last_name, leadint_dots, names)

    def alias(self) -> Optional[str]:
        """Try to read an alias. Return None if no alias is given.
        alias := (AS IDENTIFIER)?
        """
        if self.current_token.type is TokenType.as_:
            self.eat()
            value = self.current_token.value
            self.eat(TokenType.identifier)
            return value
        return None

    def import_stmt(self):
        """import_stmt := IMPORT module_meta alias"""
        pos = self.current_pos
        self.eat(TokenType.import_)
        meta = self.module_meta()
        alias = self.alias()
        return Import(meta, alias, **pos)

    def from_import_stmt(self):
        """
        id_alias := IDENTIFIER alias
        from_import_stmt := FROM module_meta IMPORT (
            STAR
            | list_of{id_alias}
            | paren_list_of{LPAREN, id_alias, RPAREN}
        )
        """
        pos = self.current_pos
        self.eat(TokenType.from_)
        meta = self.module_meta()
        self.eat(TokenType.import_)
        if self.current_token.type is TokenType.star:
            self.eat()
            return FromImportAll(meta, **pos)
        names = []
        aliases = []
        def _id_alias():
            names.append(self.current_token.value)
            self.eat(TokenType.identifier)
            aliases.append(self.alias())
        if self.current_token.type is TokenType.lparen:
            self.paren_list_of(_id_alias)
        else:
            self.list_of(_id_alias)
        return FromImport(meta, names, aliases, **pos)

    def for_stmt(self):
        """
        for_stmt := FOR IDENTIFIER IN expr COLON statement_block
        """
        pos = self.current_pos
        self.eat(TokenType.for_)
        name = self.current_token.value
        self.eat(TokenType.identifier)
        self.eat(TokenType.in_)
        expr = self.expr()
        self.eat(TokenType.colon)
        body = self.statement_block()
        return For(name, expr, body, **pos)

    def _struct_body(self):
        pos = self.current_pos
        if self.current_token.type is TokenType.identifier:
            name = self.current_token.value
            self.eat()
            self.eat(TokenType.colon)
            type_ = self.type_spec()
            res = StructField(name, type_, **pos)
        else:
            res = self.pass_stmt()
        self.eat(TokenType.new_line)
        return res

    def struct_stmt(self):
        """
        struct_field_decl := IDENTIFIER COLON type_spec
        struct_body := ((struct_field_decl | pass_stmt) NEW_LINE)*
        struct_stmt := STRUCT IDENTIFIER extends_clause
            COLON block{struct_body}
        """
        pos = self.current_pos
        self.eat(TokenType.struct)
        name = self.current_token.value
        self.eat(TokenType.identifier)
        bases = self._extends_clause()
        self.eat(TokenType.colon)
        body = self._block(self._struct_body)
        return StructDef(name, bases, body, **pos)

    def _constant_decl(self) -> Tuple[str, Optional[TypeSpec], Expression]:
        """constant_decl := IDENTIFIER (COLON type_spec)? EQUAL expr"""
        name = self.current_token.value
        self.eat(TokenType.identifier)
        if self.current_token.type is TokenType.colon:
            self.eat()
            type_ = self.type_spec()
        else:
            type_ = None
        self.eat(TokenType.equal)
        value = self.expr()
        return (name, type_, value)

    def _handle_const(self):
        """
        constant_def_body := list_of{constant_decl} |
            paren_list_of{LPAREN, constant_decl, RPAREN}
        const_statement := const_def_stmt
            | (CONST constant_def_body NEW_LINE)
        """
        self.peek()
        if self.next_token.type is TokenType.def_:
            return self.const_def_stmt()
        pos = self.current_pos
        self.eat()  # eat CONST
        names = []
        types = []
        values = []
        def _add_const_decl():
            name, type_, value = self._constant_decl()
            names.append(name)
            types.append(type_)
            values.append(value)
        if self.current_token.type is TokenType.lparen:
            self.paren_list_of(_add_const_decl)
        else:
            self.list_of(_add_const_decl)
        self.eat(TokenType.new_line)
        return ConstDef(names, types, values, **pos)

    def reference_def_stmt(self):
        """reference_def_stmt := AMPERSAND constant_decl"""
        pos = self.current_pos
        self.eat(TokenType.ampersand)
        return ReferenceDef(*self._constant_decl(), **pos)

    def result_stmt(self):
        """result_stmt := RESULT expr"""
        pos = self.current_pos
        self.eat(TokenType.result)
        return Result(self.expr(), **pos)

    def simple_new_stmt(self):
        """simple_new_stmt := NEW call_table"""
        pos = self.current_pos
        self.eat(TokenType.new)
        return NewCall(None, self.call_table(), **pos)

    def statement(self):
        """
        expr_statement := (
            var_def_stmt | auto_var_def_stmt | assign_stmt |
            aug_assign_stmt | expr_stmt | expr_new_stmt
        )
        simple_statement := (
            pass_stmt | command_stmt | import_stmt | from_import_stmt |
            result_stmt | simple_new_stmt
        )
        embedded_statement := (
            if_stmt | while_stmt | for_stmt | interface_stmt |
            def_stmt | entity_stmt | struct_stmt | inline_def_stmt
        )
        statement := ((simple_statement | expr_statement) NEW_LINE) |
            embedded_statement | const_statement
        """
        # Statements that start with special token
        TOK2STMT_EMBEDDED = {
            TokenType.if_: self.if_stmt,
            TokenType.while_: self.while_stmt,
            TokenType.interface: self.interface_stmt,
            TokenType.def_: self.def_stmt,
            TokenType.inline: self.inline_def_stmt,
            TokenType.entity: self.entity_stmt,
            TokenType.for_: self.for_stmt,
            TokenType.struct: self.struct_stmt
        }
        TOK2STMT_SIMPLE = {
            TokenType.pass_: self.pass_stmt,
            TokenType.command_begin: self.command_stmt,
            TokenType.import_: self.import_stmt,
            TokenType.from_: self.from_import_stmt,
            TokenType.ampersand: self.reference_def_stmt,
            TokenType.result: self.result_stmt,
            TokenType.new: self.simple_new_stmt
        }
        if self.current_token.type is TokenType.const:
            return self._handle_const()
        stmt_method = TOK2STMT_SIMPLE.get(self.current_token.type)
        if stmt_method:
            res = stmt_method()
            self.eat(TokenType.new_line)
            return res
        stmt_method = TOK2STMT_EMBEDDED.get(self.current_token.type)
        if stmt_method:
            return stmt_method()

        # Other statements that start with an expression
        pos = self.current_pos
        expr = self.expr()
        AUG_ASSIGN = {
            TokenType.plus_equal: Operator.add,
            TokenType.minus_equal: Operator.minus,
            TokenType.times_equal: Operator.multiply,
            TokenType.divide_equal: Operator.divide,
            TokenType.mod_equal: Operator.mod
        }

        if self.current_token.type is TokenType.equal:
            # assign_stmt := expr EQUAL expr
            self.eat()  # eat equal
            node = Assign(expr, self.expr(), **pos)
        elif self.current_token.type in AUG_ASSIGN:
            # aug_assign_stmt := expr (PLUS_EQUAL | MINUS_EQUAL
            #     | TIMES_EQUAL | DIVIDE_EQUAL | MOD_EQUAL) expr
            operator = AUG_ASSIGN[self.current_token.type]
            self.eat()  # eat operator
            node = AugmentedAssign(expr, operator, self.expr(), **pos)
        elif self.current_token.type is TokenType.colon:
            # var_def_stmt := identifier COLON type_spec (EQUAL expr)?
            self.eat()  # eat colon
            if not isinstance(expr, Identifier):
                self.error(ErrorType.INVALID_VARDEF_STMT, **pos)
            type_ = self.type_spec()
            if self.current_token.type is TokenType.equal:
                self.eat()  # eat equal
                value = self.expr()
            else:
                value = None
            node = VarDef(expr.name, type_, value, **pos)
        elif self.current_token.type is TokenType.walrus:
            # auto_var_def_stmt := identifier WALRUS expr
            self.eat()  # eat walrus
            if not isinstance(expr, Identifier):
                self.error(ErrorType.INVALID_VARDEF_STMT, **pos)
            node = AutoVarDef(expr.name, self.expr(), **pos)
        else:  # just an expr, or a new statement
            # expr_stmt := expr
            # expr_new_stmt := expr POINT NEW call_table
            node = None
            if self.current_token.type is TokenType.point:
                self.peek()
                if self.next_token.type is TokenType.new:
                    self.eat()
                    self.eat()
                    node = NewCall(expr, self.call_table(), **pos)
            if node is None:
                node = ExprStatement(expr, **pos)
        self.eat(TokenType.new_line)
        return node

    ## Other generators

    def module(self):
        """module := statement* END_MARKER"""
        pos = self.current_pos
        stmts = []
        while self.current_token.type != TokenType.end_marker:
            stmts.append(self.statement())
        return Module(stmts, **pos)

    def argument_table(self, allowed_ports: Tuple[FuncPortType],
                       type_required=True):
        """
        type_decl := COLON type_spec
        default_decl := EQUAL expr
        arg_normal := IDENTIFIER
            ((type_decl | default_decl) | (type_decl default_decl))
        arg_inline := func_port_inline IDENTIFIER type_decl?
            default_decl?
        arg_const := IDENTIFIER type_decl? default_decl?
        argtable_normal := paren_list_of{LPAREN, arg_normal, RPAREN}
        argtable_inline := paren_list_of{LPAREN, arg_inline, RPAREN}
        argtable_const := paren_list_of{LPAREN, arg_const, RPAREN}
        """
        arg_table = ArgumentTable(**self.current_pos)
        got_default = False
        def _arg_decl():
            nonlocal got_default
            pos = self.current_pos
            # read qualifier
            qualifier = self._func_port_qualifier(allowed_ports)
            # read name
            name = self.current_token.value
            self.eat()  # eat identifier
            # read type
            type_ = None
            if self.current_token.type is TokenType.colon:
                self.eat()
                type_ = self.type_spec()
            port = FunctionPort(type_, qualifier, **pos)
            # read default
            default = None
            if self.current_token.type is TokenType.equal:
                self.eat()
                default = self.expr()
                got_default = True
            elif got_default:
                self.error(ErrorType.NONDEFAULT_ARG_AFTER_DEFAULT, **pos)
            # check
            if (not (port.type or default)) and type_required:
                self.error(ErrorType.DONT_KNOW_ARG_TYPE, **pos, arg=name)
            if name in arg_table.args:
                self.error(ErrorType.DUPLICATE_ARG_DEF, **pos, arg=name)
            # add arg
            arg_table.add_arg(name, port, default)
        self.paren_list_of(_arg_decl)
        return arg_table

    def type_spec(self):
        """type_spec := expr"""
        return TypeSpec(self.expr(), **self.current_pos)

    def call_table(self):
        """
        arg := (IDENTIFIER EQUAL)? expr
        call_table := paren_list_of{LPAREN, arg, RPAREN}
        # Note this grammar does not show the fact that all keyword
        # arguments must be placed after positional arguments.
        """
        pos = self.current_pos
        args, keywords = [], {}
        got_keyword = False  # if a keyword argument has been read
        def _arg():
            nonlocal got_keyword
            self.peek()
            if self.next_token.type is TokenType.equal:
                # keyword
                got_keyword = True
                key = self.current_token.value
                pos = self.current_pos
                self.eat(TokenType.identifier)
                if key in keywords:  # if already exists
                    self.error(
                        ErrorType.ARG_MULTIPLE_VALUES, arg=key, **pos
                    )
                self.eat(TokenType.equal)
                keywords[key] = self.expr()
            else:  # positioned
                if got_keyword:
                    self.error(ErrorType.POSITIONED_ARG_AFTER_KEYWORD)
                args.append(self.expr())
        self.paren_list_of(_arg)
        return CallTable(args, keywords, **pos)

    def list_of(self, content: Callable[[], None]):
        """list_of{g} := g (COMMA g)*"""
        content()
        while self.current_token.type is TokenType.comma:
            self.eat()
            content()

    def paren_list_of(self, content: Callable[[], None],
                      lparen=TokenType.lparen):
        """paren_list_of{lp, g, rp} := lp (g COMMA)* g? rp"""
        self.eat(lparen)
        rp = BRACKETS[lparen]
        while self.current_token.type is not rp:
            content()
            if self.current_token.type is TokenType.comma:
                self.eat()
            else:
                break
        self.eat(rp)



"""Builtin support for positions."""

__all__ = ["PosType", "PosDataType", "Position"]

from typing import List, TYPE_CHECKING

from acaciamc.error import *
from acaciamc.tools import axe, cmethod_of, ImmutableMixin, transform_immutable
from acaciamc.constants import DEFAULT_ANCHOR, XYZ
from acaciamc.mccmdgen.datatype import DefaultDataType
from acaciamc.mccmdgen.ctexpr import CTDataType
from acaciamc.mccmdgen.expr import *
import acaciamc.mccmdgen.cmds as cmds
from . import entity as entity_module
from .types import Type
from .position_offset import PosOffsetDataType, PosOffset, CoordinateType
from .functions import BinaryFunction
from .rotation import RotDataType

if TYPE_CHECKING:
    from .rotation import Rotation
    from .entity import _EntityBase
    from acaciamc.mccmdgen.cmds import _ExecuteSubcmd

class PosDataType(DefaultDataType):
    name = "Pos"

ctdt_position = CTDataType("Pos")

class PosType(Type):
    def do_init(self):
        @cmethod_of(self, "__new__")
        class _new(metaclass=axe.OverloadChopped):
            """
            Pos(entity, literal["feet", "eyes"] = "feet"):
                position of entity.
                Second argument is the anchor.
            Pos(int-literal, int-literal, int-literal):
                absolute position.
                If integer is provided for x and z, they are increased
                by 0.5, so that the position is at block center.
            """
            @axe.overload
            @axe.arg("target", entity_module.EntityDataType)
            @axe.arg("anchor", axe.LiteralString())
            def from_entity(cls, compiler, target: "_EntityBase", anchor: str):
                inst = Position()
                inst.context.append(cmds.ExecuteEnv("at", target.to_str()))
                inst.context.append(cmds.ExecuteEnv("anchored", anchor))
                return inst

            @axe.overload
            @axe.arg("target", entity_module.EntityDataType)
            def from_entity_no_anchor(cls, compiler, target: "_EntityBase"):
                return cls.from_entity(compiler, target, DEFAULT_ANCHOR)

            @axe.overload
            @axe.arg("x", axe.PosXZ())
            @axe.arg("y", axe.LiteralFloat())
            @axe.arg("z", axe.PosXZ())
            def absolute(cls, compiler, x: float, y: float, z: float):
                offset = PosOffset()
                offset.set(0, x, CoordinateType.ABSOLUTE)
                offset.set(1, y, CoordinateType.ABSOLUTE)
                offset.set(2, z, CoordinateType.ABSOLUTE)
                inst = Position()
                new_inst, cmds = (inst.attribute_table.lookup("apply")
                                  .call([offset], {}, compiler))
                return new_inst, cmds

    def datatype_hook(self):
        return PosDataType()

    def cdatatype_hook(self):
        return ctdt_position

class Position(ConstExprCombined, ImmutableMixin):
    cdata_type = ctdt_position

    def __init__(self):
        super().__init__(PosDataType())
        self.context: List["_ExecuteSubcmd"] = []

        @cmethod_of(self, "dim")
        @axe.chop
        @axe.arg("id", axe.LiteralString(), rename="id_")
        @transform_immutable(self)
        def _dim(self: Position, compiler, id_: str):
            """.dim(id: str): change dimension of position."""
            self.context.append(cmds.ExecuteEnv("in", id_))
            return self
        _abs = self._create_offset_alias("abs")
        _offset = self._create_offset_alias("offset")
        """.abs(...) .offset(...)
        Alias of .apply(Offset().abs/offset(...)).
        """
        @cmethod_of(self, "local")
        @axe.chop
        @axe.arg("rot", RotDataType)
        @axe.arg("left", axe.LiteralFloat(), default=0.0)
        @axe.arg("up", axe.LiteralFloat(), default=0.0)
        @axe.arg("front", axe.LiteralFloat(), default=0.0)
        @transform_immutable(self)
        def _local(self: Position, compiler, rot: "Rotation", left: float,
                   up: float, front: float):
            """.local(rot: Rot, left: float = 0.0, up: float = 0.0,
                      front: float = 0.0)
            Apply local offset to current position, using given rotation.
            """
            offset = PosOffset.local(left, up, front)
            self.context.extend(rot.context)
            self.context.append(cmds.ExecuteEnv("positioned", str(offset)))
            return self
        @cmethod_of(self, "apply")
        @axe.chop
        @axe.arg("offset", PosOffsetDataType)
        @transform_immutable(self)
        def _apply(self: Position, compiler, offset: PosOffset):
            """.apply(offset: Offset): Apply offset to current position."""
            self.context.append(cmds.ExecuteEnv("positioned", str(offset)))
            return self
        @cmethod_of(self, "align")
        @axe.chop
        @axe.arg("axis", axe.LiteralString(), default="xyz")
        @transform_immutable(self)
        def _align(self: Position, compiler, axis: str):
            """
            .align(axis: str = "xyz"): round position on given axis down
            to the nearest integer. The axis must be a combination of
            "x", "y" and "z". For example,
            `Pos(10.5, 9.8, 10.5).align("xz")` gives `Pos(10, 9.8, 10)`.
            """
            axis_set = set(axis)
            if len(axis) != len(axis_set) or not axis_set.issubset(XYZ):
                raise Error(ErrorType.INVALID_POS_ALIGN, align=axis)
            self.context.append(cmds.ExecuteEnv("align", axis))
            return self
        self.attribute_table.set("abs", BinaryFunction(_abs))
        self.attribute_table.set("offset", BinaryFunction(_offset))

    def _create_offset_alias(self, method: str):
        @transform_immutable(self)
        def _offset_alias(self: Position, compiler, args, kwds):
            offset, cmds = (
                PosOffset().attribute_table.lookup(method)
                .call(args, kwds, compiler)
            )
            new_obj, _cmds = (self.attribute_table.lookup("apply")
                              .call([offset], {}, compiler))
            cmds.extend(_cmds)
            return new_obj, cmds
        return _offset_alias

    def copy(self) -> "Position":
        res = Position()
        res.context = self.context.copy()
        return res



__all__ = ['CTObj', 'CTDataType', 'CTObjPtr', 'CTCallable', 'CTExpr']

from typing import (
    TYPE_CHECKING, Union, List, Dict, Hashable, Iterable, Optional
)
from abc import ABCMeta, abstractmethod

from acaciamc.mccmdgen.symbol import SymbolTable
from acaciamc.mccmdgen.utils import InvalidOpError
from acaciamc.error import traced_call
from acaciamc.localization import localize

if TYPE_CHECKING:
    from acaciamc.ast import Operator
    from acaciamc.error import SourceLocation
    from acaciamc.compiler import Compiler
    from acaciamc.mccmdgen.expr import ConstExpr

class CTDataType:
    def __init__(self, name: str, bases: Iterable["CTDataType"] = ()):
        self.name = name
        self.bases = tuple(bases)

    def is_typeof(self, other: "CTExpr") -> bool:
        return self.is_baseof(abs(other).cdata_type)

    def is_baseof(self, other: "CTDataType") -> bool:
        if self is other:
            return True
        for base in other.bases:
            if self.is_baseof(base):
                return True
        return False

class CTObj:
    cdata_type: CTDataType

    def __init__(self, *args, **kwds):
        super().__init__(*args, **kwds)
        self.attributes = SymbolTable()

    def __abs__(self):
        return self

    def to_rt(self) -> "ConstExpr":
        raise InvalidOpError

    def cimplicitcast(self, type_: "CTDataType") -> "CTExpr":
        raise InvalidOpError

    def cadd(self, other: "CTObj") -> "CTExpr":
        raise InvalidOpError
    def csub(self, other: "CTObj") -> "CTExpr":
        raise InvalidOpError
    def cmul(self, other: "CTObj") -> "CTExpr":
        raise InvalidOpError
    def cdiv(self, other: "CTObj") -> "CTExpr":
        raise InvalidOpError
    def cmod(self, other: "CTObj") -> "CTExpr":
        raise InvalidOpError

    def cradd(self, other: "CTObj") -> "CTExpr":
        raise InvalidOpError
    def crsub(self, other: "CTObj") -> "CTExpr":
        raise InvalidOpError
    def crmul(self, other: "CTObj") -> "CTExpr":
        raise InvalidOpError
    def crdiv(self, other: "CTObj") -> "CTExpr":
        raise InvalidOpError
    def crmod(self, other: "CTObj") -> "CTExpr":
        raise InvalidOpError

    def cunarypos(self) -> "CTExpr":
        raise InvalidOpError
    def cunaryneg(self) -> "CTExpr":
        raise InvalidOpError
    def cunarynot(self) -> "CTExpr":
        raise InvalidOpError

    def ccompare(self, op: "Operator", other: "CTObj") -> bool:
        raise InvalidOpError

    def cdatatype_hook(self) -> CTDataType:
        raise InvalidOpError

    def cstringify(self) -> str:
        raise InvalidOpError

    def citerate(self) -> List["CTExpr"]:
        raise InvalidOpError

    def chash(self) -> Hashable:
        raise InvalidOpError

class CTObjPtr:
    def __init__(self, value: CTObj):
        self.set(value)

    def __abs__(self):
        return self.v

    def set(self, value: CTObj):
        self.v = value

CTExpr = Union[CTObj, CTObjPtr]

class CTCallable(CTObj, metaclass=ABCMeta):
    def __init__(self, *args, **kwds):
        super().__init__(*args, **kwds)
        self.source: Optional["SourceLocation"] = None
        self.func_repr = localize("objects.function.unknown")

    @abstractmethod
    def ccall(self, args: List["CTObj"], kwds: Dict[str, "CTObj"],
              compiler: "Compiler") -> CTExpr:
        pass

    def ccall_withframe(
        self, args: List["CTObj"], kwds: Dict[str, "CTObj"],
        compiler, location: Optional["SourceLocation"] = None
    ) -> CTExpr:
        return traced_call(
            self.ccall, location, self.source, self.func_repr,
            args, kwds, compiler
        )



"""Acacia symbol table."""

__all__ = ['CTRTConversionError', 'SymbolTable']

from typing import TYPE_CHECKING, Optional, Dict, Set, Union, Iterable

from acaciamc.mccmdgen import expr
from acaciamc.mccmdgen.utils import InvalidOpError

if TYPE_CHECKING:
    from acaciamc.mccmdgen.ctexpr import CTExpr
    from acaciamc.mccmdgen.expr import AcaciaExpr

class CTRTConversionError(Exception):
    def __init__(self, obj: Union["AcaciaExpr", "CTExpr"]):
        super().__init__(obj)
        self.expr = obj

class SymbolTable:
    def __init__(self, outer: Optional["SymbolTable"] = None,
                 builtins: Optional["SymbolTable"] = None):
        """
        Create a new empty symbol table.
        `outer` is looked up if a name is not found in this table.
        `builtins` is looked up finally if a name is not found in this
        table and all the outer ones. It is expected that `builtins` is
        the same for all instances of `SymbolTable`.
        """
        self.outer = outer
        self.builtins = builtins
        self.no_export: Set[str] = set()
        self._table: Dict[str, Union["AcaciaExpr", "CTExpr"]] = {}

    def set(self, name: str, value: Union["AcaciaExpr", "CTExpr"]):
        """Change value at `name` to `value`.
        If name does not exists, create it."""
        self._table[name] = value

    def delete(self, name: str):
        """Delete `name`."""
        del self._table[name]

    def update(self, d: Dict[str, Union["AcaciaExpr", "CTExpr"]]):
        """Update symbol table with `d`."""
        self._table.update(d)

    def lookup(self, name: str, use_builtins=True, use_outer=True):
        """Look up a name; if not found, return None."""
        if name in self._table:
            res = self._table[name]
            if isinstance(res, expr.AcaciaExpr):
                return res
            try:
                return abs(res).to_rt()
            except InvalidOpError:
                raise CTRTConversionError(abs(res))
        if use_outer and self.outer:
            res = self.outer.lookup(name, use_builtins=False)
            if res is not None:
                return res
        if use_builtins and self.builtins:
            res = self.builtins.lookup(name)
            if res is not None:
                return res
        return None

    def clookup(self, name: str, use_builtins=True, use_outer=True):
        if name in self._table:
            res = self._table[name]
            if isinstance(res, expr.ConstExpr):
                return res.to_ctexpr()
            elif isinstance(res, expr.AcaciaExpr):
                raise CTRTConversionError(res)
            return res
        if use_outer and self.outer:
            res = self.outer.clookup(name, use_builtins=False)
            if res is not None:
                return res
        if use_builtins and self.builtins:
            res = self.builtins.clookup(name)
            if res is not None:
                return res
        return None

    def all_names(self) -> Iterable[str]:
        return self._table.keys()

    def get_raw(self, name: str) -> Optional[Union["AcaciaExpr", "CTExpr"]]:
        return self._table.get(name)



"""All the core objects in Acacia are defined here."""

from .types import *
from .functions import *
from .module import *
from .boolean import *
from .integer import *
from .string import *
from .none import *
from .entity_template import *
from .entity import *
from .float_ import *
from .position import *
from .position_offset import *
from .rotation import *
from .entity_group import *
from .entity_filter import *
from .list_ import *
from .map_ import *
from .struct_template import *
from .struct import *
from .position_absolute import *
from .generic import *
from .extern_egroup import *



"""Acacia compiler's internal representation of types of expressions."""

__all__ = ['DataType', 'DefaultDataType', 'Storable', 'SupportsEntityField']

from typing import TYPE_CHECKING, Type as PythonType
from abc import ABCMeta, abstractmethod

if TYPE_CHECKING:
    from acaciamc.objects.entity import _EntityBase
    from acaciamc.mccmdgen.expr import AcaciaExpr, VarValue
    from acaciamc.compiler import Compiler

class DataType(metaclass=ABCMeta):
    """
    Type of Acacia expressions, like `int` or entity with a specified
    template.
    WHAT'S THE FIFFERENCE BETWEEN `Type` AND THIS?
     `Type` is just a dummy expression that implements `datatype_hook`
     so it can be used as a type specifier by users. `DataType` is not
     an expression but an internal representation of expression type. It
     specifies the type of an expression including extra information
     like template for entity and `Engroup`.
    """
    @abstractmethod
    def __str__(self) -> str:
        pass

    @classmethod
    @abstractmethod
    def name_no_generic(cls) -> str:
        pass

    @abstractmethod
    def matches(self, other: "DataType") -> bool:
        """Return whether `other` is compatible to be assigned to
        variables of this type. (i.e. if `other` is entity, check
        whether the template of `other` is a sub-template of this
        type's template).
        """
        pass

    def is_type_of(self, expr: "AcaciaExpr") -> bool:
        """Return whether `expr` is of this type."""
        return self.matches(expr.data_type)

    @classmethod
    def matches_cls(cls, other: PythonType["DataType"]) -> bool:
        """Return whether this class is of `other` type."""
        return issubclass(cls, other)

class Storable(DataType):
    """A "storable" data type. (See `AcaciaExpr`)."""
    @abstractmethod
    def new_var(self, compiler: "Compiler") -> "VarValue":
        """Construct a `VarValue` of this type."""
        pass

class SupportsEntityField(DataType):
    """Data type that can be used as a field of an entity."""
    @abstractmethod
    def new_entity_field(self, compiler: "Compiler") -> dict:
        """When a field of an entity template is this type, this is
        called. This should return a dict of data (which is called
        "field meta") that identifies this field application
        (See `IntDataType` for an example).
        If this is not defined, this type can't be used as fields.
        `new_var_as_field` should be used together with this.
        """
        pass

    @abstractmethod
    def new_var_as_field(self, entity: "_EntityBase", **meta) -> "VarValue":
        """See `new_entity_field`.
        The `meta` are the same as field meta, and this should
        return the field specified by field meta for `entity`.
        """
        pass

class DefaultDataType(DataType):
    """A data type that carries no extra information like entity
    template.
    """
    @property
    @abstractmethod
    def name(self) -> str:
        pass

    @classmethod
    def name_no_generic(cls) -> str:
        return cls.name

    def __str__(self) -> str:
        return self.name

    def matches(self, other: "DataType") -> bool:
        return (isinstance(other, DefaultDataType)
                and issubclass(type(other), type(self)))



"""Abstraction of Minecraft commands."""

from abc import ABCMeta, abstractmethod
from enum import Enum
from typing import List, NamedTuple, Optional, Union, Iterable, Callable, Dict
import json

from acaciamc.constants import TERMINATOR_CHARS

def mc_str(s: str) -> str:
    if not s:
        return '""'
    if any(char in TERMINATOR_CHARS for char in s):
        return '"%s"' % "".join(
            "\\\\" if char == "\\" else
            "\\\"" if char == '"' else
            char
            for char in s
        )
    else:
        return s

def mc_selector(s: str) -> str:
    if s.startswith("@"):
        return s
    else:
        return mc_str(s)

def mc_wc_selector(s: str) -> str:
    if s == "*":
        return s
    else:
        return mc_selector(s)

class ScbSlot(NamedTuple):
    target: str
    objective: str

    def to_str(self) -> str:
        return "%s %s" % (
            mc_wc_selector(self.target),
            mc_str(self.objective)
        )

class Command(metaclass=ABCMeta):

    is_debug = False  # only write when -d is set

    @abstractmethod
    def resolve(self) -> str:
        pass

    def func_ref(self) -> Optional["MCFunctionFile"]:
        return None

    def scb_did_read(self, slot: ScbSlot) -> bool:
        """Did read the slot or not?"""
        return False

    def scb_did_assign(self, slot: ScbSlot) -> bool:
        """Did write to the slot or not?"""
        return False

    def __repr__(self) -> str:
        return "<%s %r>" % (type(self).__name__, self.resolve())

class Cmd(Command):
    def __init__(self, cmd: str, suppress_special_cmd=False):
        self.value = cmd
        # Read command name
        if cmd and not suppress_special_cmd:
            for i, char in enumerate(cmd):
                if char in TERMINATOR_CHARS:
                    break
            name = cmd[:i]
            if name in ("scoreboard", "schedule", "execute", "function",
                        "tellraw", "titleraw"):
                raise ValueError(
                    "/%s command need to be invoked by special class" % name
                )

    def resolve(self) -> str:
        return self.value

class ScbSetConst(Command):
    def __init__(self, target: ScbSlot, value: int):
        self.target = target
        self.value = value

    def resolve(self) -> str:
        return "scoreboard players set %s %d" % (
            self.target.to_str(), self.value
        )

    def scb_did_assign(self, slot: ScbSlot) -> bool:
        return slot == self.target

class ScbAddConst(Command):
    def __init__(self, target: ScbSlot, value: int):
        self.target = target
        self.value = value

    def resolve(self) -> str:
        return "scoreboard players add %s %d" % (
            self.target.to_str(), self.value
        )

    def scb_did_assign(self, slot: ScbSlot) -> bool:
        return slot == self.target

class ScbRemoveConst(Command):
    def __init__(self, target: ScbSlot, value: int):
        self.target = target
        self.value = value

    def resolve(self) -> str:
        return "scoreboard players remove %s %d" % (
            self.target.to_str(), self.value
        )

    def scb_did_assign(self, slot: ScbSlot) -> bool:
        return slot == self.target

class ScbOp(Enum):
    ADD_EQ = "+="
    SUB_EQ = "-="
    MUL_EQ = "*="
    DIV_EQ = "/="
    MOD_EQ = "%="
    SWAP = "><"
    MIN = "<"
    MAX = ">"
    ASSIGN = "="

class ScbOperation(Command):
    def __init__(self, op: ScbOp, operand1: ScbSlot, operand2: ScbSlot):
        self.operator = op
        self.operand1 = operand1
        self.operand2 = operand2

    def resolve(self) -> str:
        return "scoreboard players operation %s %s %s" % (
            self.operand1.to_str(),
            self.operator.value,
            self.operand2.to_str()
        )

    def scb_did_assign(self, slot: ScbSlot) -> bool:
        if self.operator is ScbOp.SWAP:
            return slot == self.operand1 or slot == self.operand2
        else:
            return slot == self.operand1

    def scb_did_read(self, slot: ScbSlot) -> bool:
        if self.operator is ScbOp.SWAP:
            return slot == self.operand1 or slot == self.operand2
        else:
            return slot == self.operand2

class ScbRandom(Command):
    def __init__(self, target: ScbSlot, min_: int, max_: int):
        super().__init__()
        self.target = target
        self.min = min_
        self.max = max_

    def resolve(self) -> str:
        return "scoreboard players random %s %d %d" % (
            self.target.to_str(), self.min, self.max
        )

    def scb_did_assign(self, slot: ScbSlot) -> bool:
        return slot == self.target

class ScbObjAdd(Command):
    def __init__(self, name: str, display_name: Optional[str] = None):
        super().__init__()
        self.name = name
        self.display_name = display_name

    def resolve(self) -> str:
        if self.display_name is None:
            suffix = ""
        else:
            suffix = " %s" % mc_str(self.display_name)
        return "scoreboard objectives add %s dummy%s" % (
            mc_str(self.name), suffix
        )

class ScbObjRemove(Command):
    def __init__(self, name: str):
        super().__init__()
        self.name = name

    def resolve(self) -> str:
        return "scoreboard objectives remove %s" % mc_str(self.name)

class ScbObjDisplay(Command):
    def __init__(self, location: str, name: Optional[str],
                 order: Optional[str] = None):
        super().__init__()
        if location not in ("sidebar", "list", "belowname"):
            raise ValueError("Invalid location: %s" % location)
        if order is not None and name is not None:
            raise ValueError("Can't specify order when clearing display")
        if order is not None and location == "belowname":
            raise ValueError("Can't specify order for belowname")
        self.name = name
        self.location = location
        self.order = order

    def resolve(self) -> str:
        order = " %s" % self.order if self.order else ""
        scb = "" if self.name is None else " %s" % mc_str(self.name)
        return "scoreboard objectives setdisplay %s%s%s" % (
            mc_str(self.location), scb, order
        )

class MCFunctionFile:
    """Represents a .mcfunction file."""
    def __init__(self, path: Optional[str] = None):
        """path: full path to file."""
        self.commands: List[Command] = []
        self.set_path(path)

    def __repr__(self) -> str:
        return "<MCFunctionFile path=%r>" % self._path

    def has_content(self):
        """Return if there are any commands in this file."""
        return any((not isinstance(cmd, Comment)) for cmd in self.commands)

    def cmd_length(self):
        """Return the length of commands (not including comments)."""
        return sum(
            1
            for cmd in self.commands
            if not isinstance(cmd, Comment)
        )

    # --- About Path ---

    def get_path(self):
        if self._path is None:
            raise ValueError('"path" attribute is not set yet')
        return self._path

    def set_path(self, path: str):
        self._path = path

    def is_path_set(self) -> bool:
        return self._path is not None

    # --- Export Methods ---

    def to_str(self, debugging=False) -> str:
        return '\n'.join([
            cmd.resolve() for cmd in self.commands
            if not cmd.is_debug or debugging
        ])

    # --- Write Methods ---

    def write(self, *commands: Union[str, Command]):
        self.extend(commands)

    def write_debug(self, *comments: str):
        self.commands.extend(map(Comment, comments))

    def extend(self, commands: Iterable[Union[str, Command]]):
        for command in commands:
            if isinstance(command, Command):
                self.commands.append(command)
            else:
                self.commands.append(Cmd(command))

class Comment(Command):
    def __init__(self, comment: str, debug=True):
        if not comment.startswith("#"):
            raise ValueError("Comment must start with '#': %r" % comment)
        self.comment = comment
        self.is_debug = debug

    def resolve(self) -> str:
        return self.comment

def _scb_check_for_invoke(func: Callable[["_InvokeFunction", ScbSlot], bool]):
    # Decorator for `_InvokeFunction`.
    def _wrapped(self: "_InvokeFunction", slot: ScbSlot) -> bool:
        if self.file in self._inside_scb_check:
            return False
        self._inside_scb_check.append(self.file)
        result = func(self, slot)
        self._inside_scb_check.pop()
        return result
    return _wrapped

class _InvokeFunction(Command):
    def __init__(self, file: "MCFunctionFile"):
        super().__init__()
        self.file = file

    _inside_scb_check: List[MCFunctionFile] = []

    def func_ref(self) -> "MCFunctionFile":
        return self.file

    @_scb_check_for_invoke
    def scb_did_assign(self, slot: ScbSlot) -> bool:
        return any(cmd.scb_did_assign(slot) for cmd in self.file.commands)

    @_scb_check_for_invoke
    def scb_did_read(self, slot: ScbSlot) -> bool:
        for cmd in self.file.commands:
            if cmd.scb_did_read(slot):
                return True
        return False

class InvokeFunction(_InvokeFunction):
    def resolve(self) -> str:
        return "function %s" % self.file.get_path()

class ScheduleFunction(_InvokeFunction):
    def __init__(self, file: "MCFunctionFile", args: str):
        super().__init__(file)
        self.args = args

    def resolve(self) -> str:
        return "schedule %s %s" % (self.args, self.file.get_path())

class ScbCompareOp(Enum):
    EQ = "="
    LT = "<"
    GT = ">"
    LTE = "<="
    GTE = ">="

class _ExecuteSubcmd(metaclass=ABCMeta):
    @abstractmethod
    def resolve(self) -> str:
        pass

    def scb_did_read(self, slot: ScbSlot) -> bool:
        return False

class ExecuteEnv(_ExecuteSubcmd):
    def __init__(self, cmd: str, args: str):
        if cmd not in ("align", "anchored", "as", "at", "facing",
                       "in", "positioned", "rotated"):
            raise ValueError("Invalid env subcommand: %s" % cmd)
        self.cmd = cmd
        self.args = args

    def resolve(self) -> str:
        return "%s %s" % (self.cmd, self.args)

class ExecuteCond(_ExecuteSubcmd):
    def __init__(self, cond: str, args: str, invert=False):
        if cond not in ("entity", "block", "blocks"):
            raise ValueError("Invalid condition: %s" % cond)
        self.cond = cond
        self.args = args
        self.invert = invert

    def resolve(self) -> str:
        return " ".join(
            ("unless" if self.invert else "if", self.cond, self.args)
        )

class ExecuteScoreComp(_ExecuteSubcmd):
    def __init__(self, operand1: ScbSlot, operand2: ScbSlot,
                 operator: ScbCompareOp, invert=False):
        super().__init__()
        self.operand1 = operand1
        self.operand2 = operand2
        self.operator = operator
        self.invert = invert

    def scb_did_read(self, slot: ScbSlot) -> bool:
        return slot == self.operand1 or slot == self.operand2

    def resolve(self) -> str:
        return "%s score %s %s %s" % (
            "unless" if self.invert else "if",
            self.operand1.to_str(),
            self.operator.value,
            self.operand2.to_str()
        )

class ExecuteScoreMatch(_ExecuteSubcmd):
    def __init__(self, operand: ScbSlot, range_: str, invert=False):
        super().__init__()
        self.operand = operand
        self.range = range_
        self.invert = invert

    def scb_did_read(self, slot: ScbSlot) -> bool:
        return slot == self.operand

    def resolve(self) -> str:
        return "%s score %s matches %s" % (
            "unless" if self.invert else "if",
            self.operand.to_str(),
            self.range
        )

class Execute(Command):
    def __init__(self, subcmds: List[_ExecuteSubcmd],
                 runs: Union[Command, str]):
        # An execute without a "run" subcommand is useless in
        # mcfunctions.
        super().__init__()
        if isinstance(runs, str):
            runs = Cmd(runs)
        self.subcmds: List[_ExecuteSubcmd] = []
        for subcmd in subcmds:
            self.add_subcmd(subcmd)
        if isinstance(runs, Execute):
            self.runs = runs.runs
            for subcmd in runs.subcmds:
                self.add_subcmd(subcmd)
        else:
            self.runs = runs

    def add_subcmd(self, subcmd: _ExecuteSubcmd):
        self.subcmds.append(subcmd)

    def resolve(self) -> str:
        if not self.subcmds:
            return self.runs.resolve()
        return "execute %s run %s" % (
            " ".join(sub.resolve() for sub in self.subcmds),
            self.runs.resolve()
        )

    def scb_did_read(self, slot: ScbSlot) -> bool:
        if self.runs.scb_did_read(slot):
            return True
        for subcmd in self.subcmds:
            if subcmd.scb_did_read(slot):
                return True
        return False

    def scb_did_assign(self, slot: ScbSlot) -> bool:
        return self.runs.scb_did_assign(slot)

    def func_ref(self) -> Optional["MCFunctionFile"]:
        return self.runs.func_ref()

def execute(subcmds: List[_ExecuteSubcmd], runs: Union[Command, str]):
    """`Execute` factory. This prevents /execute being added when
    `runs` is a comment.
    """
    if isinstance(runs, Comment):
        return runs
    if isinstance(runs, str):
        if runs.lstrip().startswith("#"):
            return Comment(runs)
    return Execute(subcmds, runs)

class RawtextComponent(metaclass=ABCMeta):
    """
    An abstract class that represents a JSON rawtext component.
    This class is meant to be immutable, i.e. the result of `resolve`
    should always be the same for a given instance.
    """

    @abstractmethod
    def resolve(self) -> dict:
        pass

class RawtextText(RawtextComponent):
    def __init__(self, text: str):
        self.text = text

    def resolve(self) -> dict:
        return {"text": self.text}

class RawtextTranslate(RawtextComponent):
    def __init__(self, value: str,
                 args: Union[None, "Rawtext", List[str]] = None):
        self.value = value
        self.args = args

    def resolve(self) -> dict:
        res = {"translate": self.value}
        if isinstance(self.args, Rawtext):
            res["with"] = self.args.resolve()
        elif isinstance(self.args, list):
            res["with"] = self.args
        return res

class RawtextScore(RawtextComponent):
    def __init__(self, slot: ScbSlot):
        self.slot = slot

    def resolve(self) -> dict:
        return {"score": {"name": self.slot.target,
                          "objective": self.slot.objective}}

class RawtextSelector(RawtextComponent):
    def __init__(self, selector: str):
        self.selector = selector

    def resolve(self) -> dict:
        return {"selector": self.selector}

class Rawtext(List[RawtextComponent]):
    """An list subclass that represents a JSON rawtext."""

    def resolve(self) -> dict:
        """Convert to JSON. Caller owns the result."""
        # Optimize: Merge adjacent text components.
        res = []
        texts = []
        def _dump():
            if texts:
                s = "".join(texts)
                if s:
                    res.append(RawtextText(s).resolve())
                texts.clear()
        for c in self:
            if isinstance(c, RawtextText):
                texts.append(c.text)
            else:
                _dump()
                res.append(c.resolve())
        _dump()
        return {"rawtext": res}

class RawtextOutput(Command):
    def __init__(self, prefix: str, rawtext: Rawtext):
        self.prefix = prefix
        self.rawtext = rawtext
        # Check what score slots are used in the rawtext.
        self.score_slots: List[ScbSlot] = []
        def _visit(rawtext: Rawtext):
            for c in rawtext:
                if isinstance(c, RawtextScore):
                    self.score_slots.append(c.slot)
                elif (isinstance(c, RawtextTranslate)
                        and isinstance(c.args, Rawtext)):
                    _visit(c.args)
        _visit(self.rawtext)

    def resolve(self) -> str:
        return "%s %s" % (
            self.prefix, json.dumps(self.rawtext.resolve())
        )

    def scb_did_read(self, slot: ScbSlot) -> bool:
        return slot in self.score_slots

class TitlerawTimes(Command):
    def __init__(self, player: str, fade_in: int, stay: int, fade_out: int):
        self.player = player
        self.fade_in = fade_in
        self.stay = stay
        self.fade_out = fade_out

    def resolve(self) -> str:
        return "titleraw %s times %d %d %d" % (
            self.player, self.fade_in, self.stay, self.fade_out
        )

class TitlerawResetTimes(Command):
    def __init__(self, player: str):
        self.player = player

    def resolve(self) -> str:
        return "titleraw %s reset" % self.player

class TitlerawClear(Command):
    def __init__(self, player: str):
        self.player = player

    def resolve(self) -> str:
        return "titleraw %s clear" % self.player

class FunctionsManager:
    EXTRA_OBJ = "%s{id}"

    def __init__(self, scoreboard: str):
        self.scoreboard = scoreboard
        self.files: List["MCFunctionFile"] = []
        self._alloc_id = 0
        self._int_consts: Dict[int, ScbSlot] = {}
        self._scb_id = 0
        self._extra_obj = self.EXTRA_OBJ % scoreboard
        self.default_scb = self.add_scoreboard()

    def generate_init(self) -> List[Command]:
        res = []
        # Scoreboards
        res.append(Comment('# Register scoreboard(s)'))
        res.extend([
            ScbObjAdd(self._extra_obj.format(id=i))
            for i in range(1, self._scb_id + 1)
        ])
        # Constants
        if self._int_consts:
            res.append(Comment('# Load constants'))
            res.extend([
                ScbSetConst(slot, num)
                for num, slot in self._int_consts.items()
            ])
        return res

    def allocate(self) -> ScbSlot:
        self._alloc_id += 1
        return ScbSlot("acacia%d" % self._alloc_id, self.default_scb)

    def add_file(self, file: "MCFunctionFile"):
        self.files.append(file)

    def int_const(self, number: int) -> ScbSlot:
        if number not in self._int_consts:
            self._int_consts[number] = self.allocate()
        return self._int_consts[number]

    def add_scoreboard(self) -> str:
        self._scb_id += 1
        return self._extra_obj.format(id=self._scb_id)



"""Entity group -- a group of entities."""

__all__ = ["EGroupDataType", "EGroupGeneric", "EGroupType", "EntityGroup"]

from typing import TYPE_CHECKING, List

from acaciamc.tools import axe, resultlib, method_of, cmethod_of
from acaciamc.mccmdgen.mcselector import MCSelector
from acaciamc.mccmdgen.datatype import Storable
from acaciamc.mccmdgen.expr import *
from acaciamc.mccmdgen.utils import InvalidOpError
import acaciamc.mccmdgen.cmds as cmds
from .entity_template import ETemplateDataType
from .entity_filter import EFilterDataType
from .entity import EntityDataType, EntityReference
from .boolean import WildBool
from .integer import IntOpGroup, IntOp
from .functions import BinaryFunction, ConstructorFunction
from .generic import BinaryGeneric
from .types import TypeDataType, ctdt_type

if TYPE_CHECKING:
    from acaciamc.compiler import Compiler
    from acaciamc.mccmdgen.datatype import DataType
    from .integer import IntVar
    from .entity_template import EntityTemplate
    from .entity_filter import EntityFilter
    from .entity import _EntityBase

class EGroupDataType(Storable):
    name = "Engroup"

    def __init__(self, template: "EntityTemplate"):
        super().__init__()
        self.template = template

    def __str__(self) -> str:
        return "Engroup[%s]" % self.template.name

    @classmethod
    def name_no_generic(cls) -> str:
        return "Engroup"

    def matches(self, other: "DataType") -> bool:
        return (isinstance(other, EGroupDataType) and
                other.template.is_subtemplate_of(self.template))

    def new_var(self, compiler):
        return EntityGroup.from_template(self.template, compiler)

class EGroupGeneric(BinaryGeneric):
    def __init__(self):
        super().__init__()
        @cmethod_of(self, "__getitem__")
        @axe.chop
        @axe.arg("template", ETemplateDataType)
        def _getitem(compiler, template: "EntityTemplate"):
            return EGroupType(template)

class EGroupType(ConstExprCombined, ConstructorFunction):
    cdata_type = ctdt_type

    def __init__(self, template: "EntityTemplate"):
        super().__init__(TypeDataType())
        self.template = template

    def initialize(
        self, instance: "EntityGroup", compiler: "Compiler",
        args: "ARGS_T", keywords: "KEYWORDS_T"
    ) -> CMDLIST_T:
        @axe.chop
        def _call_me(compiler: "Compiler"):
            return resultlib.commands(instance.clear())
        _, c = BinaryFunction(_call_me).call(args, keywords, compiler)
        return c

    def datatype_hook(self):
        return EGroupDataType(self.template)

class IntEntityCount(IntOp):
    init = True

    def __init__(self, entity: str) -> None:
        self.entity = entity

    def resolve(self, var: "IntVar") -> CMDLIST_T:
        return [
            cmds.ScbSetConst(var.slot, 0),
            cmds.Execute(
                [cmds.ExecuteEnv("as", self.entity)],
                runs=cmds.ScbAddConst(var.slot, 1)
            )
        ]

class EntityGroup(VarValue):
    def __init__(self, data_type: EGroupDataType, compiler: "Compiler"):
        super().__init__(data_type)
        self.template = data_type.template
        self.tag = compiler.allocate_entity_tag()
        SELF = self.get_selector().to_str()
        MEMBER_TYPE = EntityDataType(self.template)
        OPERAND_TYPE = EGroupDataType(self.template)

        @method_of(self, "select")
        @axe.chop
        @axe.arg("filter", EFilterDataType, rename="filter_")
        def _select(compiler, filter_: "EntityFilter"):
            """
            .select(filter: Enfilter) -> EntityGroup
            Selects entities from all entities in the world that match
            the filter and add them to this entity group.
            """
            cmds = filter_.dump("tag {selected} add %s" % self.tag)
            return self, cmds
        @method_of(self, "drop")
        @axe.chop
        @axe.arg("filter", EFilterDataType, rename="filter_")
        def _drop(compiler, filter_: "EntityFilter"):
            """
            .drop(filter: Enfilter) -> EntityGroup
            Selects entities from this entity group that match the
            filter and remove them.
            """
            cmds = filter_.dump(
                "tag {selected} remove %s" % self.tag,
                among_tag=self.tag
            )
            return self, cmds
        @method_of(self, "filter")
        @axe.chop
        @axe.arg("filter", EFilterDataType, rename="filter_")
        def _filter(compiler: "Compiler", filter_: "EntityFilter"):
            """
            .filter(filter: Enfilter) -> EntityGroup
            Selects entities from this entity group that match the
            filter and only keep them.
            """
            tmp = compiler.allocate_entity_tag()
            cmds = filter_.dump(
                "tag {selected} add %s" % tmp,
                among_tag=self.tag
            )
            cmds.append("tag @e[tag={0},tag=!{1}] remove {0}".format(
                self.tag, tmp
            ))
            cmds.append("tag @e[tag={0}] remove {0}".format(tmp))
            return self, cmds
        @method_of(self, "extend")
        @axe.chop
        @axe.arg("other", OPERAND_TYPE)
        def _extend(compiler, other: "EntityGroup"):
            return self, ["tag @e[tag=%s] add %s" % (other.tag, self.tag)]
        @method_of(self, "subtract")
        @axe.chop
        @axe.arg("other", OPERAND_TYPE)
        def _subtract(compiler, other: "EntityGroup"):
            return self, ["tag @e[tag=%s] remove %s" % (other.tag, self.tag)]
        @method_of(self, "intersect")
        @axe.chop
        @axe.arg("other", OPERAND_TYPE)
        def _intersect(compiler, other: "EntityGroup"):
            return self, ["tag @e[tag=!%s] remove %s" % (other.tag, self.tag)]
        @method_of(self, "copy")
        @axe.chop
        def _copy(compiler):
            res = data_type.new_var(compiler)
            return res, self.export(res, compiler)
        @method_of(self, "clear")
        @axe.chop
        def _clear(compiler):
            return self, ["tag %s remove %s" % (SELF, self.tag)]
        @method_of(self, "add")
        @axe.chop
        @axe.star_arg("entities", MEMBER_TYPE)
        def _add(compiler, entities: List["_EntityBase"]):
            return self, ["tag %s add %s" % (entity, self.tag)
                          for entity in entities]
        @method_of(self, "remove")
        @axe.chop
        @axe.star_arg("entities", MEMBER_TYPE)
        def _remove(compiler, entities: List["_EntityBase"]):
            return self, ["tag %s remove %s" % (entity, self.tag)
                          for entity in entities]
        @method_of(self, "is_empty")
        @axe.chop
        def _is_empty(compiler: "Compiler"):
            subcmds = [cmds.ExecuteCond("entity", SELF, invert=True)]
            return WildBool(subcmds, [])
        @method_of(self, "size")
        @axe.chop
        def _size(compiler: "Compiler"):
            return IntOpGroup(init=IntEntityCount(SELF))
        @method_of(self, "to_single")
        @axe.chop
        def _to_single(compiler: "Compiler"):
            return EntityReference(self.get_selector(), self.template)
        @method_of(self, "has")
        @axe.chop
        @axe.arg("ent", MEMBER_TYPE)
        @axe.slash
        def _has(compiler: "Compiler", ent: "_EntityBase"):
            selector = ent.get_selector()
            selector.tag(self.tag)
            subcmds = [cmds.ExecuteCond("entity", selector.to_str())]
            return WildBool(subcmds, [])

    @classmethod
    def from_template(cls, template: "EntityTemplate", compiler: "Compiler"):
        return cls(EGroupDataType(template), compiler)

    def export(self, var: "EntityGroup", compiler) -> CMDLIST_T:
        # Check self-assignment
        if var.tag == self.tag:
            return []
        commands = var.clear()
        commands.append("tag @e[tag=%s] add %s" % (self.tag, var.tag))
        return commands

    def swap(self, other: "EntityGroup", compiler) -> CMDLIST_T:
        # Check self-swapping
        if other.tag == self.tag:
            return []
        return super().swap(other, compiler)

    def get_selector(self) -> "MCSelector":
        res = MCSelector("e")
        res.tag(self.tag)
        return res

    def clear(self) -> CMDLIST_T:
        return ["tag @e[tag={0}] remove {0}".format(self.tag)]

    def iadd(self, other, compiler):
        if isinstance(other, EntityGroup):
            expr, cmds = self.attribute_table.lookup("extend").call(
                [other], {}, compiler
            )
            assert expr is self
            return cmds
        raise InvalidOpError

    def isub(self, other, compiler):
        if isinstance(other, EntityGroup):
            expr, cmds = self.attribute_table.lookup("subtract").call(
                [other], {}, compiler
            )
            assert expr is self
            return cmds
        raise InvalidOpError



"""
Entity template of Acacia.

Method override rule:
    Y: ok, X: error
            Simple  Virtual  Static  NotDefined    [Method in base]
    (none)    Y        X        X        Y
    static    X        X        Y        Y
    virtual   X        X        X        Y
    override  X        Y        X        X
    [Qualifier]

Attributes can only be defined once.
Methods may be defined multiple times, if any of the following is true:
    - all of them are static, or
    - all of them are non-static, non-virtual and non-override, or
    - exactly one is virtual and all others override it.
Attributes and methods cannot have the same name.
"""

__all__ = [
    "ETemplateDataType", "EntityTemplate",
    "AcaciaNewFunction", "DEFAULT_ENTITY_NEW",
]

from typing import (
    List, Tuple, Dict, Union, Optional, Callable, NamedTuple, TYPE_CHECKING
)
from itertools import chain

from acaciamc.ast import MethodQualifier
from acaciamc.error import *
from acaciamc.tools import axe, resultlib
from acaciamc.mccmdgen import cmds
from acaciamc.mccmdgen.datatype import DefaultDataType, Storable
from acaciamc.mccmdgen.ctexpr import CTDataType
from acaciamc.mccmdgen.expr import *
from acaciamc.mccmdgen.utils import unreachable
from .entity import TaggedEntity, EntityDataType
from .functions import (
    BoundVirtualMethod, BoundMethod, InlineFunction, ConstructorFunction,
    BinaryFunction
)
from .position import PosDataType
from .none import NoneDataType
from .integer import IntVar, IntLiteral

if TYPE_CHECKING:
    from acaciamc.compiler import Compiler
    from acaciamc.mccmdgen.datatype import SupportsEntityField
    from .functions import METHODDEF_T, AcaciaFunction
    from .entity import _EntityBase
    from .position import Position

SUMMON_AT = "@a[c=1]"  # this always selects one player
# Note that @p only selects players that are alive.
SUMMON_Y = -75  # must be below -64 so it's not reachable

class ETemplateDataType(DefaultDataType):
    name = 'entity_template'

ctdt_etemplate = CTDataType("entity_template")

class _MethodDispatcher:
    """
    Every virtual method and methods that override it share one
    `_MethodDispatcher`.
    """
    def __init__(self, method_name: str, compiler: "Compiler"):
        self.compiler = compiler
        self.method_name = method_name
        self.bound: List[BoundVirtualMethod] = []
        self.impls: Dict[
            "METHODDEF_T",
            Tuple[List["EntityTemplate"], Optional[Callable[[], TaggedEntity]]]
        ] = {}
        self.result_var: Union[VarValue, None] = None

    def register(self, template: "EntityTemplate",
                 implementation: "METHODDEF_T"):
        res_type = implementation.result_type
        assert isinstance(res_type, Storable)
        if self.result_var is None:
            self.result_var = res_type.new_var(self.compiler)
        else:
            if not self.result_var.data_type.matches(res_type):
                raise Error(
                    ErrorType.OVERRIDE_RESULT_MISMATCH,
                    got=str(res_type), expect=str(self.result_var.data_type),
                    name=implementation.name
                )
        if isinstance(implementation, InlineFunction):
            get_self_var = None
        else:
            _sv = None
            def get_self_var():
                nonlocal _sv
                if _sv is None:
                    _sv = TaggedEntity.new_tag(template, self.compiler)
                    _sv.is_temporary = True
                return _sv
        self.impls[implementation] = ([template], get_self_var)
        for bound in self.bound:
            bound.add_implementation(template, implementation, get_self_var)

    def register_inherit(self, template: "EntityTemplate",
                         parent: "EntityTemplate"):
        for impl, (templates, get_self) in self.impls.items():
            if parent in templates:
                templates.append(template)
                for bound in self.bound:
                    bound.add_implementation(template, impl, get_self)
                break
        else:
            unreachable("Base not found")

    def bind_to(self, entity: "_EntityBase"):
        res = BoundVirtualMethod(
            entity, self.method_name, self.result_var, self.compiler
        )
        self.bound.append(res)
        for impl, (templates, get_self_var) in self.impls.items():
            for template in templates:
                res.add_implementation(template, impl, get_self_var)
        return res

    def bind_to_cast(self, entity: "_EntityBase"):
        best = None
        mro = entity.cast_template.mro_
        for impl, (templates, get_self_var) in self.impls.items():
            try:
                i = mro.index(templates[0])
            except ValueError:
                pass
            else:
                if best is None or i < best[0]:
                    best = i, impl, get_self_var
        _, impl, get_self_var = best
        return BoundMethod(entity, self.method_name, impl, get_self_var)

class _SimpleMethod:
    """Non-virtual and non-override methods."""
    def __init__(self, name: str, implementation: "METHODDEF_T",
                 template: "EntityTemplate", compiler: "Compiler"):
        self.name = name
        self.compiler = compiler
        self.implementation = implementation
        self.template = template
        self.is_inline = isinstance(implementation, InlineFunction)
        self._self_var = None

    def get_self_var(self) -> TaggedEntity:
        # We don't compute this inside __init__ because at that time
        # the template is not fully initialized yet.
        assert not self.is_inline
        if self._self_var is None:
            self._self_var = TaggedEntity.new_tag(self.template, self.compiler)
            self._self_var.is_temporary = True
        return self._self_var

    def bind_to(self, entity: "_EntityBase"):
        return BoundMethod(
            entity, self.name, self.implementation,
            None if self.is_inline else self.get_self_var,
        )

def _check_override(implementation: "METHODDEF_T", method: str):
    """
    Check if `implementation` can be used as virtual method or
    override a virtual method.
    """
    if isinstance(implementation, InlineFunction):
        res_type = implementation.result_type
        if not isinstance(res_type, Storable):
            raise Error(ErrorType.OVERRIDE_RESULT_UNSTORABLE,
                        name=method, type_=str(res_type))

def default_entity_new(
    compiler, template_id: AcaciaExpr, tag: str, args, keywords
):
    @axe.chop
    @axe.arg("type", axe.LiteralString(), rename="e_type")
    @axe.arg("pos", PosDataType, rename="e_pos")
    @axe.arg("name", axe.Nullable(axe.LiteralString()), default=None)
    @axe.arg("event", axe.Nullable(axe.LiteralString()), default=None)
    def new_entity(
        compiler: "Compiler", e_type: str, e_pos: "Position",
        name: Optional[str], event: Optional[str]
    ):
        e_event = "*" if event is None else event
        e_name = "" if name is None else f" {cmds.mc_str(name)}"
        e_rot = " 0 0" if compiler.cfg.mc_version >= (1, 19, 70) else ""
        return resultlib.commands([
            cmds.Execute(
                [cmds.ExecuteEnv("at", SUMMON_AT)],
                f"summon {e_type} ~ {SUMMON_Y} ~{e_rot} {e_event}{e_name}"
            ),
            cmds.Execute(
                [cmds.ExecuteEnv("at", SUMMON_AT)],
                f"tag @e[x=~,y={SUMMON_Y},z=~,dx=0,dy=0,dz=0] add {tag}"
            ),
            cmds.Execute(e_pos.context, f"tp @e[tag={tag}] ~ ~ ~"),
            *template_id.export(
                IntVar(cmds.ScbSlot(
                    f"@e[tag={tag}]", compiler.etemplate_id_scb
                )),
                compiler
            )
        ])
    _, commands = BinaryFunction(new_entity).call(args, keywords, compiler)
    return commands

def get_deleted_entity_new(template_name: str):
    def _res(compiler, template_id: AcaciaExpr, tag: str, args, keywords):
        raise Error(ErrorType.CANT_CREATE_ENTITY, type_=template_name)
    return _res

class AcaciaNewFunction(NamedTuple):
    impl: "AcaciaFunction"
    self_tag: str
    template_id_var: cmds.ScbSlot

class _DefaultEntityNewType:
    pass

DEFAULT_ENTITY_NEW = _DefaultEntityNewType()

class EntityTemplate(ConstExprCombined, ConstructorFunction):
    cdata_type = ctdt_etemplate

    def __init__(self, name: str,
                 field_types: Dict[str, "SupportsEntityField"],
                 field_metas: Dict[str, dict],
                 methods: Dict[str, AcaciaCallable],
                 method_qualifiers: Dict[str, MethodQualifier],
                 method_new: Union[AcaciaNewFunction, InlineFunction,
                                   _DefaultEntityNewType, None],
                 parents: List["EntityTemplate"],
                 compiler: "Compiler",
                 source=None):
        """
        NOTE it is the responsibility of the caller to make sure that
        `field_types` does not contain a key that is also in `methods`.
        """
        super().__init__(ETemplateDataType())
        self.name = name
        self.func_repr = self.name
        if source is not None:
            self.source = source
        self.parents = parents
        self.field_types: Dict[str, "SupportsEntityField"] = field_types
        self.field_metas: Dict[str, dict] = field_metas
        # method_dispatchers: for virtual methods AND override methods
        self.method_dispatchers: Dict[str, _MethodDispatcher] = {}
        self.simple_methods: Dict[str, _SimpleMethod] = {}
        self.static_methods: Dict[str, AcaciaCallable] = {}
        self.mro_: List[EntityTemplate] = [self]  # Method Resolution Order
        # Runtime identification number
        # Mark which template an entity is using at runtime.
        # All entities managed by Acacia have a id on scoreboard
        # `compiler.etemplate_id_scb` that represents which template it
        # is using. This `runtime_id` is the number assigned to entity
        # of this template.
        self.runtime_id = compiler.allocate_etemplate_id()
        ## MRO: We use the same C3 algorithm as Python.
        merge: List[List[EntityTemplate]] = []
        for parent in self.parents:
            if parent.mro_:
                merge.append(parent.mro_.copy())
        if self.parents:
            merge.append(self.parents.copy())
        while merge:
            for ts in merge.copy():
                candidate = ts[0]
                for ts2 in merge:
                    if candidate in ts2[1:]:
                        break
                else:
                    self.mro_.append(candidate)
                    for ts3 in merge.copy():
                        if candidate in ts3:
                            ts3.remove(candidate)
                            if not ts3:
                                merge.remove(ts3)
                    break
            else:
                raise Error(ErrorType.MRO)
        ## Check attribute name conlicts
        # Attributes may appear only once
        attr_list = list(chain(
            field_types,
            chain.from_iterable(t.field_types for t in parents)
        ))
        seen = set()
        for attr in attr_list:
            if attr in seen:
                raise Error(ErrorType.EFIELD_MULTIPLE_DEFS, attr=attr)
            else:
                seen.add(attr)
        ## Inherit attributes
        for parent in parents:
            self.field_types.update(parent.field_types)
            self.field_metas.update(parent.field_metas)
        ## Check method name conflicts
        # Methods can appear multiple times, but all override/virtual
        # methods must override from the same template (same
        # `_MethodDispatcher`).
        def _check(mx1, mx2, m):
            if m in mx1 or m in mx2:
                raise Error(ErrorType.EMETHOD_MULTIPLE_DEFS, method=m)
            if m in self.field_types:
                raise Error(ErrorType.METHOD_ATTR_CONFLICT, name=m)
        # `m_*` records methods that exist in base templates.
        m_virtual: Dict[str, _MethodDispatcher] = {}
        m_simple: Dict[str, _SimpleMethod] = {}
        m_static: Dict[str, AcaciaCallable] = {}
        for parent in reversed(self.mro_[1:]):
            for method, disp in parent.method_dispatchers.items():
                _check(m_simple, m_static, method)
                if m_virtual.setdefault(method, disp) is not disp:
                    raise Error(ErrorType.MULTIPLE_VIRTUAL_METHOD,
                                method=method)
            for method, mgr in parent.simple_methods.items():
                _check(m_virtual, m_static, method)
                m_simple[method] = mgr
            for method, impl in parent.static_methods.items():
                _check(m_virtual, m_simple, method)
                m_static[method] = impl
        ## Make sure field names do not conflict with method names
        for attr in attr_list:
            if attr in m_virtual or attr in m_simple or attr in m_static:
                raise Error(ErrorType.METHOD_ATTR_CONFLICT, name=attr)
        ## Handle methods
        for method, implementation in methods.items():
            if method in self.field_types:
                raise Error(ErrorType.METHOD_ATTR_CONFLICT, name=method)
            qualifier = method_qualifiers[method]
            if method in m_virtual:
                if qualifier is not MethodQualifier.override:
                    raise Error(ErrorType.OVERRIDE_QUALIFIER,
                                got=qualifier.localized, name=method)
                _check_override(implementation, method)
                disp = parent.method_dispatchers[method]
                self.method_dispatchers[method] = disp
                disp.register(self, implementation)
                continue
            if qualifier is MethodQualifier.none:
                if method in m_static:
                    raise Error(ErrorType.INST_OVERRIDE_STATIC, name=method)
                self.simple_methods[method] = _SimpleMethod(
                    method, implementation, self, compiler
                )
            elif qualifier is MethodQualifier.static:
                if method in m_simple:
                    raise Error(ErrorType.STATIC_OVERRIDE_INST, name=method)
                self.static_methods[method] = implementation
            elif qualifier is MethodQualifier.virtual:
                if method in m_static:
                    raise Error(ErrorType.INST_OVERRIDE_STATIC, name=method)
                if method in m_simple:
                    raise Error(ErrorType.VIRTUAL_OVERRIDE_SIMPLE, name=method)
                _check_override(implementation, method)
                disp = _MethodDispatcher(method, compiler)
                self.method_dispatchers[method] = disp
                disp.register(self, implementation)
            elif qualifier is MethodQualifier.override:
                assert method not in m_virtual  # we handled it above
                raise Error(ErrorType.NOT_OVERRIDING, name=method)
            else:
                unreachable()
        ## Inherit methods
        for method, disp in m_virtual.items():
            if method not in self.method_dispatchers:
                # Inherited override methods
                self.method_dispatchers[method] = disp
                disp.register_inherit(self, parent)
        for method, mgr in m_simple.items():
            if method not in self.simple_methods:
                # Inherited simple methods
                self.simple_methods[method] = mgr
        for method, impl in m_static.items():
            if method not in self.static_methods:
                # Inherited static methods
                self.static_methods[method] = impl
        ## `new` method
        if method_new is None:
            if len(self.mro_) > 1:
                mnew = self.mro_[1].method_new
            else:
                # This template has no base
                mnew = get_deleted_entity_new(self.name)
            mnew_src = None
        elif isinstance(method_new, AcaciaNewFunction):
            def mnew(compiler, template_id: AcaciaExpr, tag: str,
                     args, keywords):
                commands: CMDLIST_T = template_id.export(
                    IntVar(method_new.template_id_var), compiler
                )
                _, c = method_new.impl.call(args, keywords, compiler)
                commands.extend(c)
                # We don't need to clear `method_new.self_tag` here
                # because this is already done at top of the mcfunction
                # (see `Generator.visit_EntityTemplateDef`).
                commands.append(f"tag @e[tag={method_new.self_tag}] add {tag}")
                return commands
            mnew_src = method_new.impl.source
        elif isinstance(method_new, _DefaultEntityNewType):
            mnew = default_entity_new
            mnew_src = None
        else:
            def mnew(compiler, template_id: AcaciaExpr, tag: str,
                     args, keywords):
                method_new.context.entity_new_data = (template_id, tag)
                s = method_new.context.self_value = TaggedEntity(tag, self)
                s.is_temporary = True
                r, c = method_new.call(args, keywords, compiler)
                if not r.data_type.matches_cls(NoneDataType):
                    raise Error(
                        ErrorType.ENTITY_NEW_RETURN_TYPE, got=r.data_type
                    )
                # For the sake of gc...
                method_new.context.entity_new_data = None
                method_new.context.self_value = None
                return c
            mnew_src = method_new.source
        # method_new: NOTE that it is the caller's responsibility to
        # clear the given tag (third parameter).
        self.method_new: Callable[
            ["Compiler", AcaciaExpr, str, ARGS_T, KEYWORDS_T],
            CMDLIST_T
        ] = mnew
        self.method_new_source = mnew_src
        ## Register static methods to attribute table
        for name, impl in self.static_methods.items():
            self.attribute_table.set(name, impl)

    def datatype_hook(self):
        return EntityDataType(self)

    def register_entity(self, entity: "_EntityBase"):
        # Register attributes to and initialize an entity
        # whose template is self.
        # Every entities MUST call their template's `register_entity`
        assert entity.template is self
        # Convert stored virtual methods into bound method of `entity`.
        if entity.cast_template is None:
            for name, disp in self.method_dispatchers.items():
                entity.attribute_table.set(name, disp.bind_to(entity))
        else:
            for name, disp in entity.cast_template.method_dispatchers.items():
                entity.attribute_table.set(name, disp.bind_to_cast(entity))
        # Convert stored simple methods into bound method of `entity`.
        if entity.cast_template is None:
            template = self
        else:
            template = entity.cast_template
        for name, mgr in template.simple_methods.items():
            entity.attribute_table.set(name, mgr.bind_to(entity))
        # Convert stored fields to attributes of `entity`.
        for name, meta in self.field_metas.items():
            type_ = self.field_types[name]
            entity.attribute_table.set(
                name, type_.new_var_as_field(entity, **meta)
            )
        # Add static methods
        for name, impl in template.static_methods.items():
            entity.attribute_table.set(name, impl)

    def initialize(self, instance: "TaggedEntity", compiler: "Compiler",
                   args, keywords):
        """
        Calling an entity template summons a new entity, the arguments
        are passed to the constructor.
        """
        return [
            *instance.clear(),
            *self.method_new(
                compiler, IntLiteral(self.runtime_id),
                instance.tag, args, keywords
            )
        ]

    def is_subtemplate_of(self, other: "EntityTemplate") -> bool:
        # Return whether `self` is a subtemplate of `other`. A
        # template itself is considered as a subtemplate of itself.
        if self is other:
            return True
        for parent in self.parents:
            if parent.is_subtemplate_of(other):
                return True
        return False



"""
Compile-time only list container.

Purpose of list and the for-in structure are both to generate
repetitive commands. Acacia's for-in is a "compile-time loop" that
instructs the compiler to generate commands for each iteration.
Therefore, many of the functions about lists only accepts literal
expressions that can be calculated in compile time.
"""

__all__ = ["ListType", "ListDataType", "AcaciaList", "CTConstList", "CTList"]

from typing import TYPE_CHECKING, List, Union, Iterable
from itertools import repeat

from .types import Type
from .integer import IntDataType, IntLiteral
from acaciamc.tools import axe, cmethod_of
from acaciamc.error import *
from acaciamc.mccmdgen.expr import *
from acaciamc.mccmdgen.datatype import DefaultDataType
from acaciamc.mccmdgen.ctexpr import CTObj, CTObjPtr, CTDataType
from acaciamc.mccmdgen.utils import InvalidOpError

if TYPE_CHECKING:
    from acaciamc.mccmdgen.ctexpr import CTExpr

class ListDataType(DefaultDataType):
    name = "const_list"

ctdt_constlist = CTDataType("const_list")
ctdt_list = CTDataType("list", (ctdt_constlist,))

class ListType(Type):
    def do_init(self):
        @cmethod_of(self, "range")
        class _range(metaclass=axe.OverloadChopped):
            @classmethod
            def _impl(cls, *args: int):
                items = [IntLiteral(i) for i in range(*args)]
                return AcaciaList(items)

            @axe.overload
            @axe.arg("stop", axe.LiteralInt())
            def stop_only(cls, compiler, stop: int):
                return cls._impl(stop)

            @axe.overload
            @axe.arg("start", axe.LiteralInt())
            @axe.arg("stop", axe.LiteralInt())
            def start_stop(cls, compiler, start: int, stop: int):
                return cls._impl(start, stop)

            @axe.overload
            @axe.arg("start", axe.LiteralInt())
            @axe.arg("stop", axe.LiteralInt())
            @axe.arg("step", axe.LiteralInt())
            def full(cls, compiler, start: int, stop: int, step: int):
                return cls._impl(start, stop, step)
        @cmethod_of(self, "__new__")
        @axe.chop
        @axe.arg("x", axe.Iterator())
        @axe.slash
        def _new(compiler, x: ITERLIST_T):
            return AcaciaList(x)
        @cmethod_of(self, "repeat")
        @axe.chop
        @axe.arg("object", axe.AnyValue(), rename="obj")
        @axe.arg("times", axe.LiteralInt())
        def _repeat(compiler, obj: AcaciaExpr, times: int):
            return AcaciaList(repeat(obj, times))
        @cmethod_of(self, "geometric")
        @axe.chop
        @axe.arg("start", axe.LiteralInt())
        @axe.arg("ratio", axe.LiteralInt())
        @axe.arg("length", axe.LiteralInt())
        def _geometric(compiler, start: int, ratio: int, length: int):
            cur = start
            res = []
            for _ in range(length):
                res.append(IntLiteral(cur))
                cur *= ratio
            return AcaciaList(res)

    def datatype_hook(self):
        return ListDataType()

    def cdatatype_hook(self):
        return ctdt_constlist

def list2ct(x: Iterable[ConstExpr]) -> List[CTObj]:
    return [i.to_ctexpr() for i in x]

class AcaciaList(ConstExpr):
    def __init__(self, items: Iterable[ConstExpr]):
        super().__init__(ListDataType())
        self.items: List[ConstExpr] = list(items)

        @cmethod_of(self, "__getitem__")
        @axe.chop
        @axe.arg("index", axe.LiteralInt())
        def _getitem(compiler, index: int):
            self._validate_index(index)
            return self.items[index]
        @cmethod_of(self, "copy")
        @axe.chop
        def _copy(compiler):
            return AcaciaList(self.items)
        @cmethod_of(self, "slice")
        class _slice(metaclass=axe.OverloadChopped):
            @classmethod
            def _impl(cls, *args: Union[int, None]):
                return AcaciaList(self.items[slice(*args)])

            @axe.overload
            @axe.arg("stop", axe.LiteralInt())
            def stop_only(cls, compiler, stop: int):
                return cls._impl(stop)

            @axe.overload
            @axe.arg("start", axe.Nullable(axe.LiteralInt()))
            @axe.arg("stop", axe.Nullable(axe.LiteralInt()))
            def start_stop(cls, compiler, start, stop):
                return cls._impl(start, stop)

            @axe.overload
            @axe.arg("start", axe.Nullable(axe.LiteralInt()))
            @axe.arg("stop", axe.Nullable(axe.LiteralInt()))
            @axe.arg("step", axe.LiteralInt())
            def full(cls, compiler, start, stop, step: int):
                return cls._impl(start, stop, step)
        @cmethod_of(self, "cycle")
        @axe.chop
        @axe.arg("times", axe.RangedLiteralInt(0, None))
        def _cycle(compiler, times: int):
            return AcaciaList(self.items * times)
        @cmethod_of(self, "size")
        @axe.chop
        def _size(compiler):
            return IntLiteral(len(self.items))

    def _validate_index(self, index: int):
        length = len(self.items)
        if not -length <= index < length:
            raise Error(ErrorType.LIST_INDEX_OUT_OF_BOUNDS,
                        length=length, index=index)

    def iterate(self) -> ITERLIST_T:
        return self.items

    def hash(self):
        return tuple(x.hash() for x in self.items)

    def add(self, other: AcaciaExpr, compiler):
        if isinstance(other, AcaciaList):
            return AcaciaList(self.items + other.items)
        raise InvalidOpError

    def mul(self, other: AcaciaExpr, compiler):
        if isinstance(other, IntLiteral):
            return AcaciaList(self.items * other.value)
        if other.data_type.matches_cls(IntDataType):
            raise Error(ErrorType.LIST_MULTIMES_NON_LITERAL)
        raise InvalidOpError

    def to_ctexpr(self):
        return CTConstList(list2ct(self.items))

class CTConstList(CTObj):
    cdata_type = ctdt_constlist

    def __init__(self, items: Iterable["CTExpr"]):
        super().__init__()
        self.ptrs = list(map(self._new_element, items))

        @cmethod_of(self, "__ct_getitem__")
        @axe.chop
        @axe.arg("index", axe.LiteralInt())
        def _getitem(compiler, index: int):
            self._validate_index(index)
            return abs(self.ptrs[index])
        @cmethod_of(self, "copy")
        @axe.chop
        def _copy(compiler):
            return CTList(self.ptrs)
        @cmethod_of(self, "slice")
        class _slice(metaclass=axe.OverloadChopped):
            @classmethod
            def _impl(cls, *args: Union[int, None]):
                return CTList(self.ptrs[slice(*args)])

            @axe.overload
            @axe.arg("stop", axe.LiteralInt())
            def stop_only(cls, compiler, stop: int):
                return cls._impl(stop)

            @axe.overload
            @axe.arg("start", axe.Nullable(axe.LiteralInt()))
            @axe.arg("stop", axe.Nullable(axe.LiteralInt()))
            def start_stop(cls, compiler, start, stop):
                return cls._impl(start, stop)

            @axe.overload
            @axe.arg("start", axe.Nullable(axe.LiteralInt()))
            @axe.arg("stop", axe.Nullable(axe.LiteralInt()))
            @axe.arg("step", axe.LiteralInt())
            def full(cls, compiler, start, stop, step: int):
                return cls._impl(start, stop, step)
        @cmethod_of(self, "cycle")
        @axe.chop
        @axe.arg("times", axe.RangedLiteralInt(0, None))
        def _cycle(compiler, times: int):
            return CTList(self.ptrs * times)
        @cmethod_of(self, "size")
        @axe.chop
        def _size(compiler):
            return IntLiteral(len(self.ptrs))

    def _new_element(self, element: "CTExpr") -> CTObjPtr:
        return CTObjPtr(abs(element))

    def _validate_index(self, index: int):
        length = len(self.ptrs)
        if not -length <= index < length:
            raise Error(ErrorType.LIST_INDEX_OUT_OF_BOUNDS,
                        length=length, index=index)

    def citerate(self):
        return self.ptrs

    def chash(self):
        return tuple(abs(x).chash() for x in self.ptrs)

    def to_rt(self):
        return AcaciaList([abs(x).to_rt() for x in self.ptrs])

class CTList(CTConstList):
    cdata_type = ctdt_list

    def __init__(self, items: Iterable["CTExpr"]):
        super().__init__(items)

        # Methods that modify the list
        @cmethod_of(self, "__ct_getitem__", runtime=False)
        @axe.chop
        @axe.arg("index", axe.LiteralInt())
        def _getitem(compiler, index: int):
            self._validate_index(index)
            return self.ptrs[index]
        @cmethod_of(self, "extend", runtime=False)
        @axe.chop
        @axe.arg("values", axe.CTIterator())
        @axe.slash
        def _extend(compiler, values: List["CTExpr"]):
            self.ptrs.extend(map(self._new_element, values))
        @cmethod_of(self, "append", runtime=False)
        @axe.chop
        @axe.arg("value", axe.Constant())
        @axe.slash
        def _append(compiler, value: CTObj):
            self.ptrs.append(self._new_element(value))
        @cmethod_of(self, "insert", runtime=False)
        @axe.chop
        @axe.arg("index", axe.LiteralInt())
        @axe.arg("value", axe.Constant())
        @axe.slash
        def _insert(compiler, index: int, value: CTObj):
            self._validate_index(index)
            self.ptrs.insert(index, self._new_element(value))
        @cmethod_of(self, "reverse", runtime=False)
        @axe.chop
        def _reverse(compiler):
            self.ptrs.reverse()
        @cmethod_of(self, "pop", runtime=False)
        @axe.chop
        @axe.arg("index", axe.LiteralInt())
        def _pop(compiler, index: int):
            self._validate_index(index)
            return self.ptrs.pop(index)



"""
Localization Support (本地化支持)

To make sure that `set_language` works well, strings returned by
`localize` should not be stored. Instead, get a new localized string
every time. Calling `localize` on module level should be avoided, too.

"代码可读性的一小步, 消除屎山的一大步" - 鲁迅 (其实是creepebucket说的)
"""

__all__ = [
    "DEFAULT_LANGUAGE", "current_language", "set_language",
    "get_mapping", "localize", "LocalizedEnum"
]

from typing import Dict
from enum import Enum
from pkgutil import get_data

lang_cache: Dict[str, Dict[str, str]] = {}
DEFAULT_LANGUAGE = _language = 'en_US'

def current_language() -> str:
    """Get current language code."""
    return _language

def set_language(lang: str) -> None:
    """Change the language."""
    global _language
    _language = lang

def get_mapping(lang: str) -> Dict[str, str]:
    """
    Get the mapping that maps from localization key to the text of the
    specified language.
    获取指定语言的本地化映射表
    """
    if lang in lang_cache:
        return lang_cache[lang]
    lang_cache[lang] = lang_file_dict = {}  # 语言映射列表
    raw = get_data(__name__, f"lang/{lang}.lang")
    for line in raw.decode("utf-8").splitlines():
        line = line.strip()
        if line.startswith("#"):  # 注释行
            continue
        if not line:  # 空行
            continue
        key, text = line.split(" = ", maxsplit=1)  # 按等号分割语言映射
        lang_file_dict[key] = text
    return lang_file_dict

def localize(key: str) -> str:
    """Get localized text under current language."""
    mapping = get_mapping(_language)
    if key in mapping:
        return mapping[key]
    return get_mapping(DEFAULT_LANGUAGE)[key]

class LocalizedEnum(Enum):
    """
    An Enum whose values are localization keys.
    An attribute `localized` is added to each enum member which is the
    localized value using default language.
    """

    @property
    def localized(self) -> str:
        return localize(self.value)
